/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2013, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function() {
dojo.provide("com.ibm.rdm.web.compare.AttributeCompareWidget");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");

dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.compare.collectionUtils");
dojo.require("com.ibm.rdm.web.compare.CompareConstants");
dojo.require("com.ibm.rdm.web.compare.compareutils");
dojo.require("com.ibm.rdm.web.compare.compareLinksUtils");
dojo.require("com.ibm.rdm.web.compare.compareTagUtils");
dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
dojo.require("com.ibm.rdm.web.server.client.utils.WrapperResourceUtil");
dojo.require("com.ibm.rdm.web.util.deferredutils");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.server.Headers");

dojo.requireLocalization("com.ibm.rdm.web.compare", "AttributeCompareWidget");

var	MimeTypeRegistry = com.ibm.rdm.web.common.MimeTypeRegistry,
	CollectionUtils = com.ibm.rdm.web.compare.collectionUtils,
	CompareConstants = com.ibm.rdm.web.compare.CompareConstants,
	CompareUtils = com.ibm.rdm.web.compare.compareutils,
	LinksUtils = com.ibm.rdm.web.compare.compareLinksUtils,
	TagUtils = com.ibm.rdm.web.compare.compareTagUtils,
	RRMTypes = com.ibm.rdm.web.model.types.RRMTypes,
	WrapperResourceUtil = com.ibm.rdm.web.server.client.utils.WrapperResourceUtil,
	DeferredUtils = com.ibm.rdm.web.util.deferredutils,
	DomUtils = com.ibm.rdm.web.util.domutils,
	Headers = com.ibm.rdm.web.server.Headers,
	IMG_TRANSPARENT = dojo.moduleUrl("com.ibm.rdm.web", "common/images/transparent.gif").toString(),
	ATTR_NATURAL_WIDTH = 'data-rdm-natural-width',
	ATTR_NATURAL_HEIGHT = 'data-rdm-natural-height';

	
function setImageSizeAttributes(image) {
	// summary:
	//		Promise handler for setting the sizes of an image after the image has been loaded.
	//		Sets attributes ATTR_NATURAL_WIDTH and ATTR_NATURAL_HEIGHT on the image.
	// image:
	//		An Image object for an image that has just been added to the page and loaded.
	// returns:
	//		The supplied image.  Possibly promised if there needs to be a further delay
	//		before the sizes are available.
	var naturalWidth = image.naturalWidth,
		naturalHeight = image.naturalHeight,
		writeSizeAttributes = function writeSizeAttributes(width, height) {
			dojo.attr(image, ATTR_NATURAL_WIDTH, width);
			dojo.attr(image, ATTR_NATURAL_HEIGHT, height);
		};
		
	if (naturalWidth && naturalHeight) {
		writeSizeAttributes(naturalWidth, naturalHeight);
		return image;
	}
	
	// Natural width and height are not available - probably IE 8.
	// Create a new image not on display, and when it loads, find its size.
	// This load should be fast as it will be from browser cache.
	var imageDeferred = new dojo.Deferred();
	var sizerImage = new Image();
	sizerImage.onload = function() {
		if (imageDeferred.fired !== 0) {
			writeSizeAttributes(sizerImage.width, sizerImage.height);
			imageDeferred.resolve(image);
		}
	};
	sizerImage.src = image.src;
	if (sizerImage.complete) {
		// image has loaded immediately.
		writeSizeAttributes(sizerImage.width, sizerImage.height);
		imageDeferred.resolve(image);
	}
	
	return imageDeferred;
}

function getObjectLength(obj) {
	// summary:
	//		Counts the number of keys in an object
	// obj:
	//		A javascript object
	// returns:
	//		The number of keys in obj
    var count = 0;
    
    if (obj) {
    
    	for(var prop in obj) {
       		count++;
    	}
    }
    return count;
}

function scaleImage(image, scaleFactor) {
	// TODO: ideally it would be nice to use these transforms, but they seem to change the image
	// size without reducing the page area occupied.  Needs more investigation, because scaling
	// this way may give better looking scaled images. 
//	if (dojo.isFF) {
//		style = {
//			"MozTransform": "scale("+scaleFactor+")",
//			"MozTransformOrigin": "0 0"
//		};
//	} else if (dojo.isWebKit) {
//		style = {
//			"WebkitTransform": "scale("+scaleFactor+")",
//			"WebkitTransformOrigin": "0 0",
//		};
//	} else if (dojo.isIE > 8) {
//		style = {
//			"msTransform": "scale("+scaleFactor+")",
//			"msTransformOrigin": "0 0"
//		};
//	} else {
		var style = {
			"height": (parseInt(dojo.attr(image, ATTR_NATURAL_HEIGHT)) * scaleFactor) + "px",
			"width": (parseInt(dojo.attr(image, ATTR_NATURAL_WIDTH)) * scaleFactor) + "px"
		};
//	}
	dojo.style(image, style);
}

function scaleWrappedResourceImages(availableWidth, images) {
	// summary:
	//		Adjust the sizes of wrapped resource images so that they fit in the available width.
	// availableWidth:
	//		The maximum width available for the images.
	// images:
	//		An array of one or two images to resize if necessary.  The images must have ATTR_NATURAL_WIDTH
	//		set on them.
	
	// calculate sizes for the blocks that contain the images.
	// Minimum width allowed for the block containing each image is one sixth of total width - so that a
	// very small image gets some space for its label/download text.
	
	
	// Adjust the total width to allow for some space between the images and rounding errors in calculations.
	availableWidth -= 30*(images.length-1) + 40;
	var minimumBlockWidth = availableWidth / 4;

	var blockWidths = [];
	var naturalWidths = dojo.map(images, function(image){
		var naturalWidth = parseInt(dojo.attr(image, ATTR_NATURAL_WIDTH));
		return naturalWidth;
	});
	if (images.length === 1) {
		blockWidths.push(availableWidth);
	} else {
		// always size the narrower image first, so that if it is very small the
		// maximum available size is used for the wider image.
		if (naturalWidths[0] > naturalWidths[1]) {
			images.reverse();
			naturalWidths.reverse();
		}

		var blockWidth1 = naturalWidths[0];
		blockWidth1 = Math.max(blockWidth1, minimumBlockWidth);
		blockWidth1 = Math.min(blockWidth1, availableWidth/2);
		
		availableWidth -= blockWidth1;
		var blockWidth2 = naturalWidths[1];
		blockWidth2 = Math.max(blockWidth2, minimumBlockWidth);
		blockWidth2 = Math.min(blockWidth2, availableWidth);
		
		blockWidths.push(blockWidth1);
		blockWidths.push(blockWidth2);
	}
	
	for (var i = 0 ; i < images.length ; i++) {
		var blockWidth = blockWidths[i];
		var naturalWidth = naturalWidths[i];
		var image = images[i];
		if (naturalWidth > blockWidth) {
			// images needs to be scaled
			var scaleFactor = blockWidth / naturalWidth;
			scaleImage(image, scaleFactor);
		}
		var block = image.parentNode;
		dojo.style(block, "width", blockWidth+'px');
	}	
}

function displayWrappedResource(tableRow, resource, configurationURI) {
	// summary:
	//		Display a wrapped resource image and corresponding download link.
	// tableRow:
	//		Row to which a new td is appended.
	// resource:
	//		The resource for which to display content.
	// configurationURI:
	//		Configuration indicating the version of the artifact to display. 
	// returns:
	//		A promise that resolves when the image is loaded, and has had
	//		ATTR_NATURAL_WIDTH and ATTR_NATURAL_HEIGHT attributes added to it.
	//		Returns 'undefined' if the resource is not of a type that can
	//		be displayed or previewed (i.e. not an image). 
	var mimeType = resource.getWrappedResourceMimeType();
	if (!MimeTypeRegistry.isImage(mimeType)) {
		// not an image - we can't do a preview of it.
		return; // undefined
	}

	var cell = dojo.create('td', {'class':'imageCell'}, tableRow),
		wrapper = dojo.create('div', {'class':'wrapperContentImageBlock'}, cell),
		image = new Image(),
		onLoadDeferred = new dojo.Deferred();
	image.onload = function() {
		// image has loaded after some delay
		if (onLoadDeferred.fired !== 0) {
			onLoadDeferred.resolve(image);
		}
	};
	image.src = WrapperResourceUtil.getWrappedResourceImageURI(resource.getWrappedResourceURI(), configurationURI);
	wrapper.appendChild(image);
	if (image.complete) {
		// image has loaded immediately.
		onLoadDeferred.resolve(image);
	}
	return onLoadDeferred.then(setImageSizeAttributes);
}
	
dojo.declare("com.ibm.rdm.web.compare.AttributeCompareWidget", [dijit._Widget, dijit._Templated], {
	
	templatePath: dojo.moduleUrl("com.ibm.rdm.web", "compare/templates/AttributeCompareWidget.html"),
	messages: dojo.i18n.getLocalization("com.ibm.rdm.web.compare", "AttributeCompareWidget"),
	imageTransparent: IMG_TRANSPARENT,
	
	// May be set on construction (parameter to constructor or in template that instantiates this widget).  If true, when there are no
	// attribute changes to display, a message is displayed so that the display area doesn't look empty and broken.
	enableNoAttributesMessage: false,
	
	// One of the DISPLAY_AS_XXX constants, indicating the type of change being displayed.
	_displayType: null,
	
	// Attach Points
	_valuesTableBody: null,
	_publicTagsImageNode: null,
	_privateTagsImageNode: null,
	_publicTagsListNode: null,
	_attributesHeaderInfo: null,
	

	// The resources to be compared, their types, and their configurations.
	_targetResource: null,
	_targetType: null,
	_targetConfiguration: null,
	_targetConfigurationName: "",
	_sourceResource: null,
	_sourceType: null,
	_sourceConfiguration: null,
	_sourceConfigurationName: "",
	
	// This is either _targetResource or _sourceResource, depending on the type of operation.  Set when the
	// resource(s) are loaded, so that there is a single resource that can be referenced and is guaranteed
	// to be set, reducing the need to constantly look at this._displayType to work out whether _targetResource or
	// sourceResource should be used. 
	_primaryResource: null,
	_primaryConfiguration: null,
	
	// The node that is displaying primary text or wrapped resource info.
	_contentValueNode: null,
	
	// The type of the resource.
	_primaryType: null,
	
	// The attribute definitions required to display the attribute values.
	_primaryAttributeDefinitions: null,
	
	// Representations of differences that have been calculated by the server (for text differences).
	// DOM node representations of the attribute, links, tags, and collection differences.  
	// A hash keyed by attrDefURI.  Each value is an object with fields "label" and "node".
	_attributeDifferences: null,
	_publicTagDifferences: null,
	_privateTagDifferences: null,
	_collectionDifferences: null,
	_linksDifferences: null,
	
	// We may want to show something differently depending on if we are in the context of a merge.
	asMerge: false,
	
	// The current width of the widget.
	_width: 100,
	
	// <tr> node in the attributes table, used for building columns in the attributes table.
	_currentAttributeRow: null,
	
	resize: function(width, height) {
		var topPadding = dojo.style(this.domNode, "paddingTop");
		dojo.style(this.domNode, {
			width: width+"px",
			height: (height - topPadding)+"px"
		});
		this._width = width;
		this._resizeContentValue(this._width);
	},
	
	_resizeContentValue: function(/* ? Integer */widgetWidth) {
		// summary:
		//		Sets the width of the content value node that holds primary text or wrapped resource images
		//		to a size appropriate for the overall widget width.

		// Depending on its content, the primary text node can push its containing table too wide, so
		// restrict its width.  The proportion allocated must match with the table proportions
		// allocated in CSS. 
		widgetWidth = this._width;
		var ptWidth = Math.floor(widgetWidth*0.8);
		if (this._contentValueNode) {
			dojo.style(this._contentValueNode, "width", ptWidth + "px");
		}
		// resize wrapped resource images if necessary
		var images = dojo.query(".wrapperContentImageBlock img", this._contentValueNode);
		if (images.length) {
			scaleWrappedResourceImages(widgetWidth, images);
		}
	},
	
	show: function(params) {
		// summary:
		//		Loads the given resources and displays them.  Can show an added resource, a
		//		removed resource or a changed resource.
		// params:
		//		hash containing the resources, configurationURIs, type, attribute definitions and displayType to
		//		be displayed.

		this._targetResource			= params.targetResource;
		this._targetType				= params.targetType;
		this._targetConfiguration		= params.targetConfiguration;
		this._targetConfigurationName = params.targetConfigurationName || "";
		this._sourceResource			= params.sourceResource;
		this._sourceType				= params.sourceType;
		this._sourceConfiguration		= params.sourceConfiguration;
		this._sourceConfigurationName	= params.sourceConfigurationName || "";
		this._primaryResource		= params.primaryResource;
		this._primaryConfiguration	= (params.primaryResource === params.targetResource) ? params.targetConfiguration : params.sourceConfiguration;
		this._primaryType			= params.primaryType;
		this._primaryAttributeDefinitions = params.primaryAttributeDefinitions;
		this._attributeDifferences	= params.attributeDifferences;
		this._collectionDifferences = params.collectionDifferences;
		this._publicTagDifferences	= params.publicTagDifferences;
		this._privateTagDifferences = params.privateTagDifferences;
		this._linksDifferences      = params.linksDifferences;
		this._displayType			= params.displayType;
		this._teamAreaDifferences   = params.teamAreaDifferences;
		this._resourceDeltas		= params.resourceDeltas;
			
		this.clear();
		
		this._displayDifferences();
	},
	
	clear: function() {
		// summary:
		//		Clear the display, ready to display a fresh comparison,
		//		and set internal variables to default values.
		dojo.empty(this._valuesTableBody);
		dojo.empty(this._publicTagsListNode);
		dojo.empty(this._privateTagsListNode);
		dojo.empty(this._linksNode);
		dojo.empty(this._contentLabelNode);
		dojo.empty(this._contentValueNode);
		dojo.empty(this._publicTagsCount);
		dojo.empty(this._privateTagsCount);
		dojo.empty(this._collectionTable);
		dojo.empty(this._linksCount);
		dojo.empty(this._attributesCount);
		dojo.empty(this._teamOwnershipDiff);
		dojo.style(this._contentAreaNode, "display", "none");
		dojo.style(this._publicTagsRow, "display", "none");
		dojo.style(this._privateTagsRow, "display", "none");
		dojo.style(this._collectionArea, "display", "none");
		dojo.style(this._teamAreaOwnershipRow, "display", "none");
		dojo.removeClass(this._attributesSectionHead, "show");
		dojo.removeClass(this._noAttributesMessage, "show");
		dojo.removeClass(this._linksSectionHead, "show");
		dojo.removeClass(this._collectionAreaHead, "show");
	},

	_displayDifferences: function() {
		// summary:
		//		Displays the differences.
		
		var componentURI = this._primaryResource.getComponentURI(),
			showEverything = (this._displayType !== CompareConstants.DISPLAY_AS_CHANGED && this._displayType !== CompareConstants.DISPLAY_AS_LINKS_CHANGED),
			publicTagInfo = TagUtils.getTagsDiff(this._publicTagDifferences, showEverything),
			privateTagInfo = TagUtils.getTagsDiff(this._privateTagDifferences, showEverything),
			diffInfo,
			linksInfo,
			countLabel,
			numberOfPublicTags      = 0,
			numberOfPrivateTags     = 0,
			numberOfLinks           = 0,
			numberOfAttributes      = 0,
			numberOfShownAttributes = 0;
		
		// Get the counts on our tags
		numberOfPublicTags = getObjectLength(this._publicTagDifferences);
		numberOfPrivateTags = getObjectLength(this._privateTagDifferences);
		
		// Only show the counts if there are any tags, including hidden ones
		if (numberOfPublicTags > 0) {
			countLabel = dojo.string.substitute(this.messages.countLabel, [publicTagInfo.count, numberOfPublicTags]);
			this._publicTagsCount.appendChild(document.createTextNode(countLabel));
		}
		
		// Now counts for the private tags
		if (numberOfPrivateTags > 0) {
			countLabel = dojo.string.substitute(this.messages.countLabel, [privateTagInfo.count, numberOfPrivateTags]);
			this._privateTagsCount.appendChild(document.createTextNode(countLabel));
		}
		
		// Add the alt text for the tags image
		dojo.attr(this._publicTagsImageNode, 'alt', this.messages.tagsImageAlt);
		dojo.attr(this._privateTagsImageNode, 'alt', this.messages.tagsImageAlt);
		dojo.removeClass(this._publicTagsImageNode, "show");
		dojo.removeClass(this._privateTagsImageNode, "show");
		
		// Display the public tags, if we've got changes or we're showing everything
		if (publicTagInfo.hasChanged || showEverything) {

			if (!publicTagInfo.isEmpty) {
				// Need to clone the DOM node, because changing tabs in module compare deletes and
				// rebuilds the DOM. 
				this._publicTagsListNode.appendChild(dojo.clone(publicTagInfo.node));
				dojo.removeClass(this._publicTagsListNode, 'j-hint');
				
				dojo.addClass(this._publicTagsImageNode, "show");
				dojo.style(this._publicTagsRow, "display", "block");
			}
		}
		
		// Display the private tags, if we've got changes or we're showing everything
		if (privateTagInfo.hasChanged || showEverything) {
			
			if (!privateTagInfo.isEmpty) {
				// Need to clone the DOM node, because changing tabs in module compare deletes and
				// rebuilds the DOM. 
				this._privateTagsListNode.appendChild(dojo.clone(privateTagInfo.node));
				dojo.removeClass(this._privateTagsListNode, 'j-hint');
				
				dojo.addClass(this._privateTagsImageNode, "show");
				dojo.style(this._privateTagsRow, "display", "block");
			}
		}
		
		// Clear the attribute row to ensure we've got an empty slate to work on
		if (this._currentAttributeRow) {
			dojo.empty(this._currentAttributeRow);
			this._currentAttributeRow = null;
		}
		
		// Display the type if necessary
		diffInfo = CompareUtils.getTypeNameDiff(this._targetType, this._sourceType, this._primaryType);
		if (showEverything || diffInfo.hasChanged) {
			this._addRow(diffInfo);
			++numberOfShownAttributes;
		}
		++numberOfAttributes;
		
		// Display description first, and then the user-defined attributes in the order 
		// they appear in the type.
		if (!showEverything && this._displayType !== CompareConstants.DISPLAY_AS_LINKS_CHANGED) {
		
			// Show the description only if it's changed
			var attrDefURI = RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Description);
			diffInfo = this._attributeDifferences[attrDefURI];
			++numberOfAttributes;
			
			if (diffInfo.hasChanged) {
				this._addRow(diffInfo, attrDefURI);
				++numberOfShownAttributes;
			}
			
			attrDefURI = RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.AlternateSpelling);
			diffInfo = this._attributeDifferences[attrDefURI];
			++numberOfAttributes;
			
			if (diffInfo.hasChanged) {
				this._addRow(diffInfo, attrDefURI);
				++numberOfShownAttributes;
			}
			
			// display changed attributes and added or removed attributes
			dojo.forEach(this._sourceType.getAttributeDefinitions(), function(attrDefURI){
				diffInfo = this._attributeDifferences[attrDefURI];
				
				if (diffInfo.hasChanged || diffInfo.isAdded) {
					this._addRow(diffInfo, attrDefURI);
					++numberOfShownAttributes;
				}
			}, this);
			dojo.forEach(this._targetType.getAttributeDefinitions(), function(attrDefURI){
				diffInfo = this._attributeDifferences[attrDefURI];
				
				if (diffInfo.isRemoved) {
					this._addRow(diffInfo, attrDefURI);
					++numberOfShownAttributes;
				}
			}, this);
			
			// Since the entire attribute list is being shown and the sourceType and targetType attributes overlap, 
			// we want the greater of these to indicate the total number of attributes.
			numberOfAttributes += Math.max(this._targetType.getAttributeDefinitions().length, this._sourceType.getAttributeDefinitions().length);

		} else if (this._displayType !== CompareConstants.DISPLAY_AS_LINKS_CHANGED) {
			// Display the description
			diffInfo = this._attributeDifferences[RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Description)];
			this._addRow(diffInfo);
			++numberOfAttributes;
			++numberOfShownAttributes;
			
			// display all attributes
			dojo.forEach(this._primaryType.getAttributeDefinitions(), function(attrDefURI){
				diffInfo = this._attributeDifferences[attrDefURI];
				this._addRow(diffInfo);
				
				// Increment our counters
				++numberOfAttributes;
				++numberOfShownAttributes;
			}, this);
		}

		// The attributes section head is hidden initially.  Display it only if there are attributes to
		// show.
		if (numberOfShownAttributes > 0) {
			dojo.addClass(this._attributesSectionHead, "show");
		} else {
			// not displaying the Attributes section label - display an alternative message if required.
			if (this.enableNoAttributesMessage) {
				dojo.addClass(this._noAttributesMessage, "show");
			}
		}
		
		// Only show the counts if there are any attributes, including hidden ones
		if (numberOfAttributes > 0) {
			countLabel = dojo.string.substitute(this.messages.countLabel, [numberOfShownAttributes, numberOfAttributes]);
			this._attributesCount.appendChild(document.createTextNode(countLabel));
		}
		
		// Show the links counts	
		linksInfo  = LinksUtils.getLinksDiff(this._linksDifferences, false, showEverything);
		
		// Find the number of total links among all link types
		for (var linkType in this._linksDifferences) {
		
			// Make sure both link directions are accounted for.  If the object "a2bLinks" or "b2aLinks"
			// is undefined, the function will filter it and simply return 0.
			numberOfLinks += getObjectLength(this._linksDifferences[linkType].a2bLinks);
			numberOfLinks += getObjectLength(this._linksDifferences[linkType].b2aLinks);
		}
		
		// Only show the counts if there are any links, including hidden ones
		if (numberOfLinks > 0) {								
			countLabel = dojo.string.substitute(this.messages.countLabel, [linksInfo.count, numberOfLinks]);
			this._linksCount.appendChild(document.createTextNode(countLabel));
		}
		
		// Display the links only if they have changed or this is added/deleted artifact.
		if (!linksInfo.isEmpty && (showEverything || linksInfo.hasChanged)) {
			this._linksNode.appendChild(dojo.clone(linksInfo.node));
			dojo.addClass(this._linksSectionHead, "show");
		}
		
		if (this._teamAreaDifferences) {
			dojo.style(this._teamAreaOwnershipRow, "display", "block");
			
			CompareUtils.getValueDiff(this._teamAreaDifferences.target,
									  this._teamAreaDifferences.source,
									  null,
									  this._teamOwnershipDiff);
		}
		
		var label;
		// Wrapped resource content
		if (dojo.isFunction(this._primaryResource.getWrappedResourceMimeType)) {
			var wrappedResourceMimeType = this._primaryResource.getWrappedResourceMimeType();
			if (wrappedResourceMimeType) {
				label = dojo.string.substitute(this.messages.labelDelimiter, [this.messages.artifactContent]);
				dojo.place(document.createTextNode(label), this._contentLabelNode, "only");
				
				var table = dojo.create('table', {'class':'contentTable'}),
					tbody = dojo.create('tbody', null, table),
					downloadLinkRow = dojo.create('tr', null, tbody),
					previewRow = dojo.create('tr'),
					imageLoadPromises = [],
					cell, linkText, imagePromise;
				if (this._displayType === CompareConstants.DISPLAY_AS_CHANGED) {
					// TODO: is there a better way to tell if a wrapped resource has changed?  Can we compare the wrappedResourceURIs?
					// build the download links
					
					cell = dojo.create('td', {'class':'linkCell'}, downloadLinkRow);
					linkText = dojo.string.substitute(this.messages.downloadFrom, [this._sourceConfigurationName]);
					cell.appendChild(WrapperResourceUtil.buildDownloadLink(linkText, this._sourceResource, this._sourceConfiguration));
					
					cell = dojo.create('td', {'class':'linkCell'}, downloadLinkRow);
					linkText = dojo.string.substitute(this.messages.downloadFrom, [this._targetConfigurationName]);
					cell.appendChild(WrapperResourceUtil.buildDownloadLink(linkText, this._targetResource, this._targetConfiguration));

					// build the preview images
					
					imagePromise = displayWrappedResource(previewRow, this._sourceResource, this._sourceConfiguration);
					if (imagePromise) {
						imageLoadPromises.push(imagePromise);
					}
					
					imagePromise = displayWrappedResource(previewRow, this._targetResource, this._targetConfiguration);
					if (imagePromise) {
						imageLoadPromises.push(imagePromise);
					}
					
				} else {
					// build the download link
					cell = dojo.create('td', {'class':'linkCell'}, downloadLinkRow);
					cell.appendChild(WrapperResourceUtil.buildDownloadLink(this.messages.downloadForView, this._primaryResource, this._primaryConfiguration));
					// build the preview image
					imagePromise = displayWrappedResource(previewRow, this._primaryResource, this._primaryConfiguration);
					if (imagePromise) {
						imageLoadPromises.push(imagePromise);
					}
				}
				
				if (imageLoadPromises.length) {
					// got a preview to display
					tbody.appendChild(previewRow);
				} else {
					// no preview - display a message - this ends up above the download links on display
					dojo.create('div', {'class':'noPreview j-hint'}, this._contentValueNode)
						.appendChild(document.createTextNode(this.messages.noPreview));
				}
				this._contentValueNode.appendChild(table);
				dojo.style(this._contentAreaNode, "display", "block");
				// When the images are loaded, resize them so that they appear side-by side.
				DeferredUtils.all(imageLoadPromises)
					.then(dojo.hitch(this, scaleWrappedResourceImages, this._width));
				// we can bail now, it can't be a wrapped resource AND have primary text etc.
				return;
			}
		}
				
		// If it's a collection, let's show its resources
		if (this._collectionDifferences) {
			// Show the header and the changes table
			dojo.addClass(this._collectionAreaHead, "show");
			diffInfo = CollectionUtils.generateCollectionDiff(this._collectionDifferences, this.asMerge, this._resourceDeltas);
			dojo.place(dojo.clone(diffInfo.node), this._collectionTable, "only");
			dojo.style(this._collectionArea, "display", "block");
		}
				
		// And finally, primary text
		var primaryTextAttrDefURI = RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.PrimaryText);
		if (this._primaryResource.hasAttribute(primaryTextAttrDefURI)) {
			var primaryTextAttrDef = this._primaryAttributeDefinitions[primaryTextAttrDefURI];
			diffInfo = this._attributeDifferences[primaryTextAttrDef.getResourceURI()];
			if (diffInfo) {
				label = dojo.string.substitute(this.messages.labelDelimiter, [diffInfo.label]);
				dojo.place(document.createTextNode(label), this._contentLabelNode, "only");	
				if (this._displayType === CompareConstants.DISPLAY_AS_LINKS_CHANGED) {
					diffInfo.node = dojo.create("div", {
						"innerHTML": this._targetResource.getPrimaryHtmlText() 
					});
				}
								
				dojo.when(diffInfo.node, dojo.hitch(this, function(node){			
					// 88402: Do not allow clicks on anchor tags. Replace them with their content only, but precede it with a hyperlink decoration.
					dojo.query("a", node).forEach(function(anchor) {
						var replacementNode = DomUtils.changeNodeTag(anchor, "span");
						dojo.create("div", {"class":"inlineTextIconNode hyperlinkImg"}, replacementNode, "first");
					});
					
					dojo.query("img", node).forEach(dojo.hitch(this, function(img) {
						if (this._displayType === CompareConstants.DISPLAY_AS_DELETED) {
							img.src += "&"+Headers.VVC_CONFIGURATION+"="+this._targetConfiguration;
						} else {
							img.src += "&"+Headers.VVC_CONFIGURATION+"="+this._sourceConfiguration;
						}
					}));
					// Clone the node before display because on IE if we try to re-display the node after it has been
					// put on display and then removed (through clear()) we find that the node has been emptied.
					dojo.place(dojo.clone(node), this._contentValueNode, "only");
					this._onPrimaryTextDiffInfo(diffInfo);
				}));
			}
			dojo.style(this._contentAreaNode, "display", "block");
		}
	},
	
	_onPrimaryTextDiffInfo: function(diffInfo) {
		// Overrideable method for use in inherited classes that is called after the diffInfo for the primary text is resolved.
		return;
	},

	_addRow: function(itemToShow) {
		// summary:
		//		Add a row for the given item
		// itemToShow:
		//		An object with a DOM "node", "label", "isEmpty" and "hasChanged" elements.

		if (!itemToShow) {
			// Nothing to display
			return;
		}
		
		// To get two columns of attributes in the table, add every second one to the row
		// created by the previous call.
		var row;
		if (this._currentAttributeRow) {
			// got a row that has one previous entry.  Use it this time (but not next time).
			row = this._currentAttributeRow;
			this._currentAttributeRow = null;
		} else {
			// start a new row
			row = this._currentAttributeRow = dojo.create("tr", null, this._valuesTableBody);
		}
		
		var nameCell = dojo.create("td", {"class":"nameCol j-label"}, row),
			valueCell = dojo.create("td", {"class":"valueCol"}, row),
			label = itemToShow.label;
		// label from itemToShow can be a DOMNode (for added or removed attribute) or a string
		if (typeof label === 'string') {
			label = document.createTextNode(label);
		}
		nameCell.appendChild(label);
		nameCell.appendChild(document.createTextNode(this.messages.labelDelimiterChar));
		// Clone the node before display because on IE if we try to re-display the node after it has been
		// put on display and then removed (through clear()) we find that the node has been emptied.
		
		// We can have either a real node or a promise of a node.  Clone and display the node only
		// after it has been fully built, else we might get an empty or partially built value.
		dojo.when(itemToShow.node, function(node){
			valueCell.appendChild(dojo.clone(node));
		});
	}
});
	
})();
