/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2014, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function() {
 
	dojo.provide("com.ibm.rdm.web.compare._ModuleCompareWidgetBase");
	
	dojo.require("com.ibm.rdm.web.compare.CompareConstants");
	dojo.require("com.ibm.rdm.web.compare.compareutils");
	dojo.require("com.ibm.rdm.web.util.domutils");
	dojo.require("com.ibm.rdm.web.common.MessageHandler");
	dojo.require("com.ibm.rdm.web.compare.compareTagUtils");
	dojo.require("com.ibm.rdm.web.compare.compareLinksUtils");
	dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
	dojo.require("com.ibm.rdm.web.server.client.utils.WrapperResourceUtil");
	dojo.require("jazz.util.ResizeNotifier");
	dojo.require("dojox.gfx.canvas");
	dojo.require("com.ibm.rdm.web.common.MimeTypes");

	dojo.requireLocalization("com.ibm.rdm.web.compare", "ModuleStructureCompareWidget");

// Summary:
//		This object provides some functions which are common between
//		ModuleStructureCompareWidget and ModuleStructureMergeWidget.
//		

	var CompareUtils = com.ibm.rdm.web.compare.compareutils;
	var CompareConstants = com.ibm.rdm.web.compare.CompareConstants;
	var TagUtils = com.ibm.rdm.web.compare.compareTagUtils;
	var DomUtils = com.ibm.rdm.web.util.domutils;
	var MessageHandler = com.ibm.rdm.web.common.MessageHandler;
  	var RRMTypes = com.ibm.rdm.web.model.types.RRMTypes;
  	var WrapperResourceUtil = com.ibm.rdm.web.server.client.utils.WrapperResourceUtil;
  	var MimeTypeRegistry = com.ibm.rdm.web.common.MimeTypeRegistry;
  	var LinksUtils = com.ibm.rdm.web.compare.compareLinksUtils;
	var ROW_URI_ATTR = "data-rdm-uri";
	var SCROLL_ARROW_HEIGHT = dojo.isIE >= 10 ? 34 : 17;
	var ROW_CONTENT_WIDTH_ADJUST_ATTR = "data-rdm-width-adjust",
        CONTENT_COL_HEADING_TAG_ATTR = "data-rdm-heading-tag",
        IMG_TAG = dojo.moduleUrl("com.ibm.rdm.web", "common/images/resources/tag.gif").toString();
	
	var messages;
	
	dojox.gfx.switchTo("canvas");
	
	function addPropertyRowIfChanged(change, wrapperNode) {
		// summary:
		//		Add a representation of an attribute or similar property to the changed attributes area
		//		of a cell.  Only adds to the display if there is a meaningful change to show. 
		// change:
		//		Change information object, including label, change node representation and hasChanged indicator.
		//		May also have isAdded and isRemoved indicators (which are considered to be changes). 
		// wrapperNode:
		//		A tbody node to which a new row is appended. 
		// TODO: get feed back from UX here, for now show all properties.
		if (change /*change.hasChanged || change.isAdded || change.isRemoved*/) {
			var rowNode = dojo.create("tr"),
				labelNode = dojo.create("td", {"class":"propertyLabel j-label"}, rowNode),
				label = change.label;
			// label can sometimes be a string and sometimes a DOM node (in the case where the
			// attribute is added or removed by a change of artifact type).
			if (typeof label === 'string') {
				label = document.createTextNode(label);
			}
			labelNode.appendChild(label);
			labelNode.appendChild(document.createTextNode(messages.labelDelimiterChar));
			var valueNode = dojo.create("td", {"class":"propertyValue"}, rowNode);
			// change.node may be a DOM node or a promise of a DOM node
			dojo.when(change.node, function(node){
				valueNode.appendChild(node);
			});
			wrapperNode.appendChild(rowNode);
		}
	}
	
	function resizeWideContentNodesForGrid(gridNode) {
		// summary:
		//		Sets the width on rich text nodes and images to match a new row width.  This permits
		//		the side-scrolling on the rich text and images to properly match the available width
		//		in the table.
		//		When they are added, the primary text nodes and images have class "wideContent" and they have
		//		a data attribute ROW_CONTENT_WIDTH_ADJUST_ATTR.  ROW_CONTENT_WIDTH_ADJUST_ATTR is a
		//		number that is the difference between the width of the table and the available width
		//		for the rich text node.  The number may be different for each row, to account for
		//		indenting of the content on the row.  The new width for the rich text area is calculated
		//		from the current row width and the ROW_CONTENT_WIDTH_ADJUST_ATTR value.
		// gridNode:
		//		DOM node for the left or right grid area.
			
		var wideContentNodes = dojo.query(".wideContent["+ROW_CONTENT_WIDTH_ADJUST_ATTR+']', gridNode);
		var firstWideNode = wideContentNodes[0];
		if (!firstWideNode) {
			// nothing to do
			return;
		}
		// Find the row width of the row containing the first wide content.  They should all be the same
		// width, so no point looking it up for each rich text node.
		var rowNode = DomUtils.findAncestorByAttr(firstWideNode, ROW_URI_ATTR, gridNode);
		if (!rowNode) {
			// give up - shouldn't ever get here!
			return;
		}
		var rowWidth = rowNode.offsetWidth;
		wideContentNodes.forEach(function(node){
			dojo.style(node, "width", (rowWidth - dojo.attr(node, ROW_CONTENT_WIDTH_ADJUST_ATTR)) + "px");
		});
	}

	
	function getAnnotationTypeFromClassName(n) {
		// Summary:
		//
		//	Get a string either "added", "removed" or "moved"
		//  if the parameter has that class.  
		// n:
		//	A DOM node which has one of "added", "removed" or "moved"
		//	class names.
		if (!n) {
			return;
		}
		
		if (dojo.hasClass(n, "added")) {
			return "added";
		} else if (dojo.hasClass(n, "removed") || dojo.hasClass(n, "removedPlaceholder")) {
			return "removed";
		} else if (dojo.hasClass(n, "moved")) {
			return "moved";
		} else if (dojo.hasClass(n, "valueChanged")) {
			return "valueChanged";
		}
	}

	dojo.declare("com.ibm.rdm.web.compare._ModuleCompareWidgetBase", null, {
		
	
		constructor: function() {
			if (!messages) {
				messages = dojo.i18n.getLocalization("com.ibm.rdm.web.compare", "ModuleStructureCompareWidget");
			}
			
			this._jazzUIResizeHandle = jazz.util.ResizeNotifier.register({
				context: this,
	  			force: true,
	  			callback: this._showAnnotations
			});
			
		},
	
	
		isRowFullyVisible: function(connectionNode, scrollParent) {
			// summary:
			//		Indicates whether the given connection node is fully visible within the view are of its scroller.
			// connectionNode:
			//		A node within a row to which a connection is potentially made - usually the same size as the row node.
			// scrollParent:
			//		The node that is a scroller around connectionNode.
			// returns:
			//		True if the row is fully visible.  Falsey if it is scrolled off screen, or is only partially visible.
			var scrollTop = scrollParent.scrollTop,
				scrollBottom = scrollTop + scrollParent.offsetHeight,
				rowTop = connectionNode.offsetTop,
				rowBottom = rowTop + connectionNode.offsetHeight;
				
			return ((rowTop >= scrollTop) && (rowBottom <= scrollBottom)); 
		},
	
		getConnectionNode: function(idOrPlaceholderNode) {
			// summary:
			//		Finds the node against which a connection should be made, given
			//		a node with an Id on it or a visible placeholder.  The node for the connection is not always
			//		the same as the Id node, because of the placeholder nodes in the tree.
			// idOrPlaceholderNode:
			//		A node which can be any node that has an id on it (a row or the id node within
			//		a placeholder) or a placeholder node (i.e. the displayed node for placeholder that contains
			//		one or more invisible nodes that have the actual placeholder IDs on them.
			// returns:
			//		An object like:
			//			{ node: connectionNode, connectionOffset: offsetPixels }
			if (dojo.hasClass(idOrPlaceholderNode, "addedPlaceholder") || dojo.hasClass(idOrPlaceholderNode, "removedPlaceholder") || dojo.hasClass(idOrPlaceholderNode, "movedPlaceholder")) {
				return {
					node: idOrPlaceholderNode,
					connectionOffset: 4
				};
			}
			var parent = idOrPlaceholderNode.parentNode;
			if (dojo.hasClass(parent, "addedPlaceholder") || dojo.hasClass(parent, "removedPlaceholder") || dojo.hasClass(parent, "movedPlaceholder")) {
				return {
					node: parent,
					connectionOffset: 4
				};
			}
			var connectionNode = idOrPlaceholderNode.firstChild;
			// Node should always have a child, but just in case, return the original if no child node.
			return {
				node: connectionNode ? connectionNode : idOrPlaceholderNode,
				connectionOffset: 8
			};
		},
		
		_scrollPane: function(referenceConnectionNode, connectionNodeToScroll, gridNode, connectionOffset) {
			// summary:
			//		Scrolls the given grid, such that the connectionNodeToScroll is in view, and is lined up
			//		with referenceConnectionNode if possible.
			// referenceConnectionNode:
			//		The row node that is staying still.
			// connectionNodeToScroll:
			//		The row node in the given grid, that is to be scrolled to match with the position of the
			//		referenceConnectionNode.
			// gridNode:
			//		The grid scroller node that is to be scrolled.
			// connectionOffset:
			//		An offset to be applied to the scrolling, to account for the fact that not all connections
			//		meet their nodes at exactly the same point.  This corrects the scroll slightly to ensure that
			//		if there is a connection line it is horizontal.
			// returns:
			//		A possibly deferred boolean, indicating that the scroll has completed.
			var referenceConnectionViewPosition = referenceConnectionNode.offsetTop - referenceConnectionNode.offsetParent.scrollTop;
			var scrollPaneTarget = connectionNodeToScroll.offsetTop - referenceConnectionViewPosition + connectionOffset;
	
			var scrollConnectHandle,
				curve = [gridNode.scrollTop, scrollPaneTarget],
				scrollDeferred;
			if (curve[0] === curve[1]) {
				// no scrolling to be done.  scrolling is done if there is no scroll in progress at the moment.
				return !this._scrollAnimation;
			} else {
				// got some scrolling to do.  Stop any previous animation and start
				// a new one.
				if (this._scrollAnimation) {
					this._scrollAnimation.stop();
					this._scrollAnimation = null;
				}
				scrollDeferred = new dojo.Deferred();
				var anim = new dojo._Animation({
					curve: curve,
					onEnd: dojo.hitch(this, function(){
						//console.log("ending");
						dojo.disconnect(scrollConnectHandle);
						this._scrollAnimation = null;
						scrollDeferred.callback(true);
					}),
					onStop: function() {
						//console.log("stopping");
						scrollDeferred.callback(false);
					}
				});
				scrollConnectHandle = dojo.connect(anim, "onAnimate", this, function(offset){
					gridNode.scrollTop = offset;
					this._showConnections();
				});
				
				this._scrollAnimation = anim;
				
				anim.play();
				return scrollDeferred;
			}
		},
		
		_scrollOppositeSelection: function(isLeftGrid, rowNode) {
			// Scroll the corresponding row on the opposite grid into view if it is not already visible.
			// _selectRow has just set this._rightSelection and this._leftSelection, so use them to identify the
			// row to be scrolled into view.
			var oppositeSelection = isLeftGrid ? this._rightSelection : this._leftSelection;
			if (oppositeSelection) {
				var selectedConnectionNodeInfo = this.getConnectionNode(rowNode);
				var oppositeSelectedConnectionNodeInfo = this.getConnectionNode(oppositeSelection.rowNode);
				// Find out if the opposite row is visible  Only force a scroll of the opposite pane if the row being highlighted there is not visible. 
				var scrollNode = isLeftGrid ? this._rightGridScroller : this._leftGridScroller,
					connectionOffset = oppositeSelectedConnectionNodeInfo.connectionOffset - selectedConnectionNodeInfo.connectionOffset;
					
				var isVisible = this.isRowFullyVisible(oppositeSelectedConnectionNodeInfo.node, scrollNode);
				
				if (!isVisible) {
					// Scroll the opposite pane to bring the corresponding row into view.
					this._scrollPane(selectedConnectionNodeInfo.node, oppositeSelectedConnectionNodeInfo.node, scrollNode, connectionOffset);
				}
			}
		},
		
		_selectRow: function(isLeftGrid, rowNode, /*?boolean*/dontTakeFocus) {
			// summary:
			//		Select a row identified by the given row node, and calls
			//		onSelect() to which our wrapper widget can connect.  The intention
			//		is that this causes display of the artifact in the row.
			//		If there is a corresponding row on the opposite grid that
			//		is also selected.
			// isLeftGrid:
			//		boolean, indicating whether the node is on the left or the right grid.
			// rowNode:
			//		DOM node of the row to be selected.  This can be a real row, or a placeholder (which is
			//		actually a node within a row node.
			// dontTakeFocus:
			//		If provided and true, the selection does not grab focus, allowing a selection to be
			//		made on initial display without grabing focus from elsewhere on the page.
			var displayType, oppositeGrid, thisGridScroller, oppositeId, oppositeRowNode, artifactURI, oppositeArtifactURI, placeholderIdNode, isPlaceholder;
			if (!rowNode) {
				return;
			}
			if (dojo.hasClass(rowNode, "added")) {
				displayType = CompareConstants.DISPLAY_AS_ADDED;
			} else if (dojo.hasClass(rowNode, "removed")) {
				displayType = CompareConstants.DISPLAY_AS_DELETED;
			} else if (dojo.hasClass(rowNode, "valueChanged")) {
				displayType = CompareConstants.DISPLAY_AS_CHANGED;
			} else if (dojo.hasClass(rowNode, "addedPlaceholder")) {
				// We've been given a placeholder.
				displayType = CompareConstants.DISPLAY_AS_ADDED;
				isPlaceholder = true;
			} else if (dojo.hasClass(rowNode, "removedPlaceholder")) {
				displayType = CompareConstants.DISPLAY_AS_DELETED;
				isPlaceholder = true;
			} else {
				displayType = CompareConstants.DISPLAY_AS_NOT_CHANGED;
			}
			
			// Find the corresponding row in the opposite grid.
			thisGridScroller = isLeftGrid ? this._leftGridScroller : this._rightGridScroller;
			oppositeGrid = isLeftGrid ? this._rightGrid : this._leftGrid;
			if (isPlaceholder) {
				// click is on a placeholder
				placeholderIdNode = rowNode.firstChild;
				oppositeId = isLeftGrid ? this._getRightIdFromLeft(placeholderIdNode.id) : this._getLeftIdFromRight(placeholderIdNode.id);
				oppositeRowNode = dojo.query("#"+oppositeId, oppositeGrid)[0];
				// Should always have an oppositeNode, but protect just in case...
				if (oppositeRowNode) {
					oppositeArtifactURI = dojo.attr(oppositeRowNode, ROW_URI_ATTR);
				}
			} else if (displayType === CompareConstants.DISPLAY_AS_CHANGED || displayType === CompareConstants.DISPLAY_AS_NOT_CHANGED) {
				// click is on a changed or not changed row
				oppositeId = isLeftGrid ? this._getRightIdFromLeft(rowNode.id) : this._getLeftIdFromRight(rowNode.id);
				oppositeRowNode = dojo.query("#"+oppositeId, oppositeGrid)[0];
				// Should always have an oppositeNode, but protect just in case...
				if (oppositeRowNode) {
					oppositeArtifactURI = artifactURI = dojo.attr(oppositeRowNode, ROW_URI_ATTR);
				}
			} else if (displayType === CompareConstants.DISPLAY_AS_ADDED || displayType === CompareConstants.DISPLAY_AS_DELETED) {
				// click is on an added or removed row
				oppositeId = isLeftGrid ? this._getRightIdFromLeft(rowNode.id, true) : this._getLeftIdFromRight(rowNode.id);
				oppositeRowNode = dojo.query("#"+oppositeId, oppositeGrid)[0];
			}
			
			artifactURI = artifactURI || dojo.attr(rowNode, ROW_URI_ATTR);
			// select the row and let others know.
			if (artifactURI || oppositeArtifactURI) {
				// remove any previous selections including removing them as browser tab stops.
				if (this._leftSelection) {
					dojo.removeClass(this._leftSelection.rowNode, "selected");
					dojo.removeAttr(this._leftSelection.highlightNode, "tabIndex");
					this._leftSelectionOverlay.hide();
					this._leftSelection = null;
				}
				if (this._rightSelection) {
					dojo.removeClass(this._rightSelection.rowNode, "selected");
					dojo.removeAttr(this._rightSelection.highlightNode, "tabIndex");
					this._rightSelectionOverlay.hide();
					this._rightSelection = null;
				}
				
				dojo.addClass(rowNode, "selected");
				var highlightNode = this._setSelectionOverlay(isLeftGrid, rowNode);
				var mainSelection = {
					rowNode: rowNode,
					highlightNode: highlightNode
				};
				// Make the new highlight node a browser tab stop
				dojo.attr(highlightNode, "tabIndex", "0");
				if (!dontTakeFocus) {
					highlightNode.focus();
				}
	
				var oppositeSelection;
				if (oppositeRowNode) {
					dojo.addClass(oppositeRowNode, "selected");
					highlightNode = this._setSelectionOverlay(!isLeftGrid, oppositeRowNode);
					oppositeSelection = {
						rowNode: oppositeRowNode,
						highlightNode: highlightNode
					};
					// Make the new highlight node a browser tab stop
					dojo.attr(highlightNode, "tabIndex", "0");
				}
				if (isLeftGrid) {
					this._leftSelection = mainSelection;
					this._rightSelection = oppositeSelection;
				} else {
					this._leftSelection = oppositeSelection;
					this._rightSelection = mainSelection;
				}
				
			this.onSelect({
					fromArtifactURI: isLeftGrid ? artifactURI : oppositeArtifactURI,
					toArtifactURI: isLeftGrid ? oppositeArtifactURI: artifactURI,
					displayType: displayType
				});
			}
		},
		
		onSelect: function() {
		},
		
		_clearAnnotations: function() {
			dojo.query(".annotation", this.domNode).forEach(function(a) { 
				a.parentNode.removeChild(a);
			});
		},
		
		_generateAnnotations: function() {
			// Summary:
			//
			//	Generates a list of annotations from the right hand grid.  These are
			//  objects of the form: 
			//  
			//		start: 		DOM node that the annotation should start on
			//		end:   		DOM node that the annotation should end on, possibly the same
			//			   		as the start.
			//		count: 		The number of rows in this annotation.
			//		getMessage: function that returns a message that could be shown
			//					on hover e.g.
			//	
			//
			
			var annotations = [],
				currentAnnotation = {},
				nl = dojo.query(".row.added, .row.valueChanged, .row.removed, .row.moved, .removedPlaceholder", this._rightGrid);

			nl.forEach(function(node) {
				var thisNodeClass    = getAnnotationTypeFromClassName(node),
					prevSiblingClass = getAnnotationTypeFromClassName(node.previousSibling),
					nextSiblingClass = getAnnotationTypeFromClassName(node.nextSibling);
						
				if (prevSiblingClass !== thisNodeClass) {
					// we need a new annotation. Either there is no
					// previous sibling, or it is not of the same type
					// as the current one.
					currentAnnotation = {
						start: node,
						type: thisNodeClass,
						count: 1,
						getMessage: function() {
							return this.count + 
								(this.count > 1 ? " rows " : " row ") + this.type;
						}
					};
					
					annotations.push(currentAnnotation);
				} else {
					currentAnnotation.count++;
				}
				
				if (thisNodeClass !== nextSiblingClass) {
					// if the next sibling either does not exist
					// or has a different type than this node, then
					// we reached the end of a block.
					currentAnnotation.end = node;
				}
			});
					
			return annotations;
		},
		
		_connectImgLoadToAnnotationRedraw: function() {
			// Summary: 
			//  Queries for img nodes and connects to the onload event, which is a cue to repaint
			//  the annotations as the height of the grid may have changed.

			dojo.query("img", this._rightGrid).connect("load", dojo.hitch(this, this._renderAnnotationsIfHeightChanged));	
		},
		
		_renderAnnotationsIfHeightChanged: function() {
			// Summary:
			//	Generates and renders the annotations if the height has changed since the
			//  last time we did it.
			if (this._lastMeasuredRightGridHeight !== this._rightGrid.scrollHeight) {
				this._showAnnotations();
			}	
		},
		
		_showAnnotations: function() {
			setTimeout(dojo.hitch(this, this._showAnnotationsImpl));
		},
		
		_showAnnotationsImpl: function() {
			var middleGridScrollHeight = this._rightGrid.scrollHeight,
				middleGridViewportHeight = this._rightGrid.offsetHeight,
				annotations = [],				
				fragment = document.createDocumentFragment(),
				annotationTopRange = middleGridViewportHeight - (SCROLL_ARROW_HEIGHT * 2);
			
			if (this._rightGrid.scrollHeight > this._rightGrid.clientHeight) {
				annotations = this._generateAnnotations();
			}
			
			// remove all the old ones...
			this._clearAnnotations();
			
			annotations.forEach(function(a) {
				
				var top = a.start.offsetTop,
					height = (a.end.offsetTop + a.end.offsetHeight) - a.start.offsetTop,
					// compute a number for roughly how far down the scrollable height
					// the start of this annotation goes
					percentage = top / middleGridScrollHeight,
					// pos is the top position of the annotation DOM node
					pos = (annotationTopRange * percentage) + SCROLL_ARROW_HEIGHT;
				
				dojo.create("div", {
					"class": "annotation " + a.type,
					"style": {
						"top": pos + "px",
						"height": annotationTopRange * (height / middleGridScrollHeight) + "px"
					},
					"title": a.getMessage(),
					"onclick": dojo.hitch(this, function() {
						this._selectRow(false, a.start);
						// try and get the middle of the selected thing in the middle of the viewport.
						this._rightGridScroller.scrollTop = ((a.start.offsetTop + height / 2) - (this._rightGridScroller.clientHeight / 2));
						this._scrollOppositeSelection(false, a.start);
					})
				}, fragment);
			}, this);
	
			// finally, add all the annotations in one go...
			this.domNode.appendChild(fragment);	
			this._lastMeasuredRightGridHeight = middleGridScrollHeight;			
		},
		
		_displayWrappedResource: function(appendToNode, rowData, configurationURI) {
			// summary:
			//		Adds a download links and an image relating to a wrapped resource.
			// appendToNode:
			//		A node within the row to which the image is to be appended.
			// rowData:
			//		Partial or full requirement for the row.  Must include a wrapped resource URI of an image type.
			// configurationURI:
			//		Configuration containing the required version of the wrapped artifact.
			
			// Images can be wide, so we need to add a content width adjustment attribute to the row so that
			// the node containing them can be resized, so giving a horizontal scrollbar on the image rather than on the
			// whole module.
			// Calculate the width offset for the row.  This is used to calculate the width of the
			// new content when first added, and upon resize.  For this we use the size of a div containing the
			// image, but we cannot measure that div until the node we return here has been added to
			// the DOM.  Therefore, just create the wrapper div here and follow up after a timeout to measure the
			// size of it and add the image itself.
			
			var resourceName = rowData.getTitle(),
				downLoadLinktext = resourceName ? dojo.string.substitute(messages.downloadByName, [resourceName]) : messages.downloadForView,
				downloadLink = WrapperResourceUtil.buildDownloadLink(downLoadLinktext, rowData, configurationURI),
				mimeType = rowData.getWrappedResourceMimeType();
			dojo.addClass(downloadLink, "downloadLink");
			appendToNode.appendChild(downloadLink);
	
			if (MimeTypeRegistry.isImage(mimeType)) {
				// display an image
				var imageWrapper = dojo.create('div', null, appendToNode);
				window.setTimeout(dojo.hitch(this, function(){
					var rowNode = DomUtils.findAncestorByAttr(imageWrapper, ROW_URI_ATTR, this.domNode);
					if (rowNode) {
						dojo.attr(imageWrapper, ROW_CONTENT_WIDTH_ADJUST_ATTR, rowNode.offsetWidth - imageWrapper.offsetWidth);
						dojo.addClass(imageWrapper, "wideContent");
					}
					// Load the image
					var img = new Image();
					img.onload = dojo.hitch(this, this._onInlineImageLoaded);
					img.src = WrapperResourceUtil.getWrappedResourceImageURI(rowData.getWrappedResourceURI(), configurationURI);
					imageWrapper.appendChild(img);
				}), 0);
			} else {
				// no image to display - display a message - this ends up above the download links on display
				dojo.create('div', {'class':'noPreview j-hint'}, downloadLink, 'before')
					.appendChild(document.createTextNode(messages.noPreview));
			}
		},
		
		// COMMON
		_setSelectionOverlay: function(isLeftGrid, /* ?DomNode */rowNode) {
			// summary:
			//		Positions the selection overlay on the given row.
			// isLeftGrid:
			//		Boolean indicating which grid the row is in.
			// rowNode:
			//		A node in the row.  If null or undefined the selection overlay is
			//		hidden.
			// returns:
			//		The node over which the overlay is positioned.  May be the supplied rowNode
			//		or a node within it.  Returns undefined if no rowNode was provided. 
			var selectionOverlay = isLeftGrid ? this._leftSelectionOverlay : this._rightSelectionOverlay;
			if (!rowNode) {
				selectionOverlay.hide();
				return;
			}
			
			// If it is in a move block, highlight the deepest nest.
			var highlightNode = dojo.query(".deepestNest", rowNode)[0];
			
			if (!highlightNode && dojo.hasClass(rowNode, "removed")) {
				highlightNode = rowNode;
			}
			
			if (!highlightNode) {
				// This covers placeholders
				highlightNode = rowNode;
			}
						
			var marginBox = dojo.marginBox(highlightNode);
			selectionOverlay.move(marginBox);
			selectionOverlay.show();
			return highlightNode;
		},
		
		// COMMON
		_redrawSelectionOverlays: function() {
			// summary:
			//		Redraws the selection overlay highlights to match any changes due
			//		to cells in the grid resizing.
			
			if (this._leftSelection) {
				this._setSelectionOverlay(true, this._leftSelection.rowNode);
			}
	
			if (this._rightSelection) {
				this._setSelectionOverlay(false, this._rightSelection.rowNode);
			}
			
		},
		
		_showChanges: function(isLeftGrid, contentColumnNode, rowNode, showChangesLinkNode) {
			// summary:
			//		Replaces the content of the given content column node with
			//		calculated changes for the artifact on a row.
			
			// If the changes have previously been displayed, and are now hidden, show them again
			var changesContentNode = dojo.query(".changesContent", contentColumnNode)[0];
			var initialContentNode = dojo.query(".initialContent", contentColumnNode)[0];
			if (changesContentNode) {
				// no need to fetch anything - just show the nodes.
				dojo.removeClass(initialContentNode, "show");
				dojo.addClass(changesContentNode, "show");
				this._redrawSelectionOverlays();
				this._showConnections();
				this._renderAnnotationsIfHeightChanged();
				return;
			}
			
			// Calculate the width offset for the row.  This is used to calculate the width of the
			// new content when first added, and upon resize.
			var widthAdjust = dojo.attr(rowNode, ROW_CONTENT_WIDTH_ADJUST_ATTR);
			if (!widthAdjust) {
				var widthNode = dojo.query(".textContent", rowNode)[0];
				if (widthNode) {
					dojo.attr(rowNode, ROW_CONTENT_WIDTH_ADJUST_ATTR, rowNode.offsetWidth - widthNode.offsetWidth);
				}
			}
			
			var displayType;
			if (dojo.hasClass(rowNode, "added")) {
				displayType = CompareConstants.DISPLAY_AS_ADDED;
			} else if (dojo.hasClass(rowNode, "removed")) {
				displayType = CompareConstants.DISPLAY_AS_DELETED;
			} else if (dojo.hasClass(rowNode, "valueChanged")) {
				displayType = CompareConstants.DISPLAY_AS_CHANGED;
			} else {
				displayType = CompareConstants.DISPLAY_AS_NOT_CHANGED;
			}
	
			// Find the corresponding row in the opposite grid.
			var oppositeGrid = isLeftGrid ? this._rightGrid : this._leftGrid;
			var oppositeId = isLeftGrid ? this._getRightIdFromLeft(rowNode.id) : this._getLeftIdFromRight(rowNode.id);
			var oppositeRowNode = dojo.query("#"+oppositeId, oppositeGrid)[0];
			// Should always have an oppositeNode, but protect just in case...
			var oppositeArtifactURI;
			if (oppositeRowNode) {
				oppositeArtifactURI = dojo.attr(oppositeRowNode, ROW_URI_ATTR);
			}
			
			var artifactURI = dojo.attr(rowNode, ROW_URI_ATTR);
					
			// get the changes and display them
			var fromArtifactURI = isLeftGrid ? artifactURI : oppositeArtifactURI,
				toArtifactURI = isLeftGrid ? oppositeArtifactURI: artifactURI;
			var loadingIndicatorNode = this._showLoadingChanges(showChangesLinkNode);
			var hideLoadingChangesFn = dojo.hitch(this, this._hideLoadingChanges, showChangesLinkNode, loadingIndicatorNode);
			CompareUtils.getArtifactChanges(displayType, fromArtifactURI, this._compareContext.targetConfiguration, toArtifactURI, this._compareContext.sourceConfiguration)
				.then(dojo.hitch(this, this._displayAttributeDifferences, contentColumnNode, initialContentNode, rowNode))
				.always(hideLoadingChangesFn)
				.otherwise(dojo.hitch(MessageHandler, MessageHandler.showMessageForError));
		},
		
		_showLoadingChanges: function(showChangesLinkNode) {
			// summary:
			//		Replace the showChanges link with a loading indicator, so that the link cannot be
			//		clicked a second time.  The link itself is not deleted, but is just hidden, so that it
			//		can be displayed again if hide changes is clicked.
			// showChangesLinkNode:
			//		The show changes link that is to be replaced on display with a loading indicator.
			// returns:
			//		The newly created loading indicator node - so that it can be given to a subsequent call
			//		to _hideLoadingChanges().
			var loadingIndicator = dojo.create('span', {'class': 'loading'});
			loadingIndicator.appendChild(document.createTextNode(messages.loading));
			dojo.style(showChangesLinkNode, "display", "none");
			dojo.place(loadingIndicator, showChangesLinkNode, "after");
			return loadingIndicator;
		},
		
		_hideLoadingChanges: function(showChangesLinkNode, loadingIndicatorNode) {
			// summary:
			//		Remove the loading indicator, and reinstate the showChanges node.  After the
			//		initial load the loading indicator is no longer needed, because subsequent
			//		show/hide changes operations just show or hide the existing nodes, without
			//		fetching the data again.  The show changes link is reinstated, but will still
			//		be hidden (due to other classes higher in the DOM) until hide changes is clicked. 
			// showChangesLinkNode:
			//		The show changes link that is to be shown again if hide changes is clicked.
			// loadingIndicatorNode:
			//		The loading indicator node that is to be removed.
			dojo.destroy(loadingIndicatorNode);
			dojo.style(showChangesLinkNode, "display", "inline");
			this._renderAnnotationsIfHeightChanged();
		},
		
		_resizeWideContentNodes: function() {
			// summary:
			//		Sets the width on rich text nodes and images to match a new row width.  This permits
			//		the side-scrolling on the rich text and images to properly match the available width
			//		in the table.
			resizeWideContentNodesForGrid(this._leftGrid);
			resizeWideContentNodesForGrid(this._rightGrid);
		},
		
		_onInlineImageLoaded: function() {
			// summary:
			//		Callback indicating that an image (e.g. a wrapped resource) has been loaded.
			
			// When an image loads, row positions shuffle, so need to redraw various things.
			
			this._resizeWideContentNodes();
			this._showConnections();
			this._redrawSelectionOverlays();
			this._renderAnnotationsIfHeightChanged();
		},
		
		_displayAttributeDifferences: function(contentColumnNode, initialContentNode, rowNode, changeInfo) {

			var changesContentNode = dojo.create('div', {"class":"displayContent changesContent show"});
	
			// display link to hide changes
			var selector = dojo.create('div', {"class":"contentSelector"}, changesContentNode);
			var link = dojo.create('a', {"class":"hideChangesLink","tabIndex":"-1"}, selector);
			link.setAttribute('href', '#');
			if (dojo.hasClass(rowNode, "valueChanged")) {
				link.appendChild(document.createTextNode(messages.hideChanges));
			} else {
				link.appendChild(document.createTextNode(messages.hideContent));
			}
			
			var componentURI = changeInfo.primaryResource.getComponentURI();
			var change, labelNode, label, valueNode, wrappedResourceMimeType, sectionHead;
			
			
			// START display changed attributes and tags
			var propertiesWrapper = dojo.create('div', {"class":"changedProperties com-ibm-rdm-web-compare"}, changesContentNode);
	
			// Display Tags
			// show all changes :
			change = TagUtils.getTagsDiff(changeInfo.publicTagDifferences, true/*false*/);
			// But only when we have tag differences
			if (change.count > 0/*.hasChanged*/) {
				var tagsRow = dojo.create('div', {'class':'tagsRow'}, propertiesWrapper);
				dojo.create('img', {'class':'tagImageLabel', src:IMG_TAG, alt:change.label}, tagsRow);
				dojo.create('div', {'class':'tagsList public'}, tagsRow).appendChild(change.node);
			}
			
			var attributesIndenter = dojo.create('div', {"class":"indentSection"});
			var attributesTable = dojo.create('table', {"class":"changedAttributes com-ibm-rdm-web-compare"}, attributesIndenter);
			var attributesTableBody = dojo.create('tbody', null, attributesTable);
			
			// display title
			change = changeInfo.attributeDifferences[RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Title)];
			if (dojo.hasClass(rowNode, "linkEndpointOnly")) {
				change.node = dojo.create("div", {
					"innerHTML": changeInfo.targetResource.getAttributeValue(RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Title))
				});
			}
			addPropertyRowIfChanged(change, attributesTableBody);
			
			// display type name
			change = CompareUtils.getTypeNameDiff(changeInfo.targetType, changeInfo.sourceType, changeInfo.primaryType);
			addPropertyRowIfChanged(change, attributesTableBody);
	
			// Display description if it is set
			change = changeInfo.attributeDifferences[RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Description)];
			if (dojo.hasClass(rowNode, "linkEndpointOnly")) {
				change.node = dojo.create("div", {
					"innerHTML": changeInfo.targetResource.getAttributeValue(RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Description))
				});
			}
			addPropertyRowIfChanged(change, attributesTableBody);
			
			// Display user-defined attributes
			dojo.forEach(changeInfo.primaryType.getAttributeDefinitions(), function(attrDefURI){
				change = changeInfo.attributeDifferences[attrDefURI];
				
				if (dojo.hasClass(rowNode, "linkEndpointOnly")) {
					change.node = dojo.create("div", {
						"innerHTML": changeInfo.targetResource.getAttributeValue(attrDefURI)
					});
				}
				
				addPropertyRowIfChanged(change, attributesTableBody);
			}, this);
			// Add in attributes that have been removed due to a change of artifact type
			if (changeInfo.targetType && changeInfo.sourceType && (changeInfo.targetType !== changeInfo.sourceType)) {
				// There is a change of type
				dojo.forEach(changeInfo.targetType.getAttributeDefinitions(), function(attrDefURI){
					change = changeInfo.attributeDifferences[attrDefURI];
					if (change.isRemoved) {
						addPropertyRowIfChanged(change, attributesTableBody);
					}
				});
			}
			
			// Only add the label and the attributes area if it contains something
			if (attributesTableBody.firstChild) {
				// A label for the attributes section
				sectionHead = dojo.create('div', {'class':'sectionHead'}, propertiesWrapper);
				dojo.create('span', {'class':'name'}, sectionHead).appendChild(document.createTextNode(messages.attributes));
				// display the table
				propertiesWrapper.appendChild(attributesIndenter);
			}
			
			// END display changed attributes
			
			// display links
			
			LinksUtils.mixinResourceDeltaInfoWithLinksDiff(changeInfo.linksDifferences, this._resourceDeltas);
			
			var linksDiff = LinksUtils.getLinksDiff(changeInfo.linksDifferences, true/*compact*/, false);
			
			
			
			if (!linksDiff.isEmpty) {
				// A label for the attributes section
				sectionHead = dojo.create('div', {'class':'sectionHead'}, propertiesWrapper);
				dojo.create('span', {'class':'name'}, sectionHead).appendChild(document.createTextNode(messages.links));
				// The links themselves
				dojo.create('div', {"class":"indentSection"}, propertiesWrapper).appendChild(linksDiff.node);
			}
					
			// display wrapped resource if there is one
			if (dojo.isFunction(changeInfo.primaryResource.getWrappedResourceMimeType)) {
				wrappedResourceMimeType = changeInfo.primaryResource.getWrappedResourceMimeType();
				if (wrappedResourceMimeType) {
					this._displayWrappedResource(changesContentNode, changeInfo.primaryResource, this._compareContext.sourceConfiguration);
				}
			}
			
			// display primary text
			var primaryTextAttrDefURI = RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.PrimaryText);
			if (changeInfo.primaryResource.hasAttribute(primaryTextAttrDefURI)) {
	
				change = changeInfo.attributeDifferences[primaryTextAttrDefURI];
				
				if (dojo.hasClass(rowNode, "linkEndpointOnly")) {
					change.node = dojo.create("div", {
						"innerHTML": changeInfo.targetResource.getPrimaryHtmlText()
					});
				}
				
				var ptContentNode = dojo.create('div', {"class":"com-ibm-rdm-web-compare wideContent rdm-text-artifact-widget"}, changesContentNode),
					ptAppendNode = ptContentNode;
				
				// wrap in a heading tag if the row has been marked as a heading type
				var headingTag = dojo.attr(contentColumnNode, CONTENT_COL_HEADING_TAG_ATTR);
				if (headingTag) {
					// heading text
					ptAppendNode = dojo.create(headingTag, null, ptContentNode);
				}
				
				// Add the primary text itself, remembering that change.node may be a DOM node or a
				// promise of a DOM node.
				dojo.when(change.node, function(node){
					ptAppendNode.appendChild(node);
				});
				
				// Pick up the width adjust value that was set on the row before changes were fetched in _showChanges.
				var widthAdjust = dojo.attr(rowNode, ROW_CONTENT_WIDTH_ADJUST_ATTR);
				var contentWidth = rowNode.offsetWidth - widthAdjust;
				dojo.style(ptContentNode, "width", contentWidth+"px");
				// Copy the attribute on to this node, so that it is easy to find on resize.
				dojo.attr(ptContentNode, ROW_CONTENT_WIDTH_ADJUST_ATTR, widthAdjust);
			}
	
			// Add our new content to the display, and hide the initial content.  We keep the initial content
			// but hide it so that it can be shown again.
			dojo.removeClass(initialContentNode, "show");
			contentColumnNode.appendChild(changesContentNode);
			this._redrawSelectionOverlays();
			this._showConnections();
	
		},
		
		_hideChanges: function(contentColumnNode) {
			// summary:
			//		Hides the changes in the given content column node, and displays the original cotent.
			var changesContentNode = dojo.query(".changesContent", contentColumnNode)[0];
			var initialContentNode = dojo.query(".initialContent", contentColumnNode)[0];
			dojo.removeClass(changesContentNode, "show");
			dojo.addClass(initialContentNode, "show");
			this._redrawSelectionOverlays();
			this._renderAnnotationsIfHeightChanged();
			this._showConnections();
		},
		
		uninitialize: function() {
			jazz.util.ResizeNotifier.unregister(this._jazzUIResizeHandle);
		}
		
	});

} ());