/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2008, 2016.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
 
 (function() {
dojo.provide("com.ibm.rdm.web.history.RevisionsWidget");
dojo.require("com.ibm.rdm.web.history.RevisionsCache");
dojo.require("com.ibm.rdm.web.common.ProjectCache");
dojo.require("com.ibm.rdm.web.history.HistoryHover");
dojo.require("com.ibm.rdm.web.model.resources.ResourceManager");
dojo.require("com.ibm.rdm.web.snapshots.ModuleSnapshotServer");
dojo.require("com.ibm.rdm.web.snapshots.ModuleSnapshotHover");

var RevisionsCache = com.ibm.rdm.web.history.RevisionsCache;
var ProjectCache = com.ibm.rdm.web.common.ProjectCache;

dojo.declare(
	"com.ibm.rdm.web.history.RevisionsWidget",
	 null,
{ 	
	///////////////////////////////////////////////////////////////////////////////////
	// Public fields

	resourceURI: null,
	componentURI: null,
	baselineURI: null,
	// function set on construction.  Called to set the page URI to match selected content.
	// parameters resourceURI, revision, baselineURI
	setPageURI: null,
	historyViewer: null,
	timeLineWidget : null,
	defaultPageSize: 25,	
	currentDateRange: null,
	dateRangePages: [],
	
	///////////////////////////////////////////////////////////////////////////////////
	// Private fields

	///////////////////////////////////////////////////////////////////////////////////
	// Public API

	constructor: function(args){			
		dojo.mixin(this,args);		
		this._revisionInfos = null;
		this._revisionInfosMap = {};
		this._initialRevision = null;
	},
	
	
	onThumbnailsInfoLoaded: function(revisionsParser) {
		this.revisions = revisionsParser.getTimeSlicedRevisionInfos();
		this._revisionInfos = revisionsParser.getRevisionInfos();
		this._revisionInfosMap = revisionsParser._revisionInfosMap;
		this._initialRevision = revisionsParser.initialRevision;
		this.timeLineWidget._timeLineDiv.style.overflow="hidden";
		this.timeLineWidget.setNewestRevision(revisionsParser.initialRevision.revisionId);
		this._processThumbNails();
		
	},
	
	getDateRange: function(revision){
		return this._getFromRevisionsInfoMap(revision).dateRange;
	},
	
	getNextPage: function(event){
		var dateRange = dojo.attr(event.currentTarget, "date-range");
		this.dateRangePages[dateRange]++;
		this.timeLineWidget.cleanTimeTableNodes();
		this._processThumbNails();
	},
	
	// TODO kludgy code that won't be long-lived once revisions of versions are supported
	_getFromRevisionsInfoMap: function(revision) {
		// summary:
		//		Look up the requested revision from _revisionInfosMap and return it.
		// This is necessary because the code currently has an inconsistency about whether a "revision" is a full revision URL
		// or just a base URL (of a version). This map normally stores full revision URLs. But, the code requesting it may only
		// know the base URL. If a lookup of the provided URL doesn't work as-is, we'll walk through the map to find a match by
		// base URL only.
		// NOTE: This'll cause problems once revisions of versions are supported; we won't have a way to distinguish the revisions
		// from each other when more than one of them has the same base version URL.

		// See if map has an exact match
		var ret = this._revisionInfosMap[revision];
		if (ret) {
			return ret;
		}
		
		// No exact match, truncate the provided revision to be a base URL
	    var i = revision.indexOf("?");
	    var revisionBaseUrl = revision;
	    if (i >= 0) {
	    	revisionBaseUrl = revision.substring(0, i);
	    }

	    // Go through the map, truncate each of the URLs and compare
		for (var key in this._revisionInfosMap) {
			if (this._revisionInfosMap.hasOwnProperty(key) ) {
			    i = key.indexOf("?");
			    var keyBaseUrl;
			    if (i >= 0) {
			    	keyBaseUrl = key.substring(0, i);
			    }
			    else {
			    	keyBaseUrl = key;
			    }
			    if (keyBaseUrl === revisionBaseUrl) {
			    	return this._revisionInfosMap[key];
			    }
			} 
		}
		
		return null;
	},
	
 	_processThumbNails: function() {
		for (var revisionsInDateRange in this.revisions) {			
			this.currentDateRange = revisionsInDateRange;
			this.dateRevisions = this.revisions[revisionsInDateRange];	
			if (this.dateRangePages[revisionsInDateRange] == null) {
				this.dateRangePages[revisionsInDateRange] = 1;	
			}
			this.timeLineWidget.onThumbnailsInfoLoaded(this.dateRevisions.slice(0, (this.defaultPageSize * this.dateRangePages[revisionsInDateRange])), this.dateRevisions.length);
		}
		
		var i = 1;
		for (var dateRange in this.dateRangePages) {			
			if (this.dateRangePages[dateRange] > 2 || i > 1) {				
				this.timeLineWidget.addScrollArrows();
				break;
			}
			i++;
		}
	},
	
	selectEntry: function(revision) {
		var revisionInfo = this._getFromRevisionsInfoMap(revision);
		if (revisionInfo) {
			var revDate = (typeof revisionInfo.updated === 'object' ?  revisionInfo.updated : dojo.date.stamp.fromISOString(revisionInfo.updated));
			var formattedDate = dojo.date.locale.format(revDate,{formatLength:"medium"});
			this.timeLineWidget.setModifiedDate(formattedDate);
		}
	},
	
	createHover: function(div, revision) {
		var historyHoverPopup;
		if(revision.id.indexOf("baseline") > 0) {
			// URI to show this module artifact in the context of the VVC configuration/baseline associated with this particular thumbnail
			var launchURI = this.resourceURI+"?vvc.configuration="+encodeURIComponent(revision.id);
			historyHoverPopup = new com.ibm.rdm.web.snapshots.ModuleSnapshotHover({baselineURI : launchURI});
			historyHoverPopup.createHoverPopup(div, revision);
		} else {
			historyHoverPopup = new com.ibm.rdm.web.history.HistoryHover({
				previousRevision: revision["http://jazz.net/xmlns/v0.6#predecessor"],
				currentRevision: revision["revisionId"],
				resourceURI: this.resourceURI,
				modifiedDate: revision.updated,
				authorURI: revision.authorURI,
				snapshotTitle: revision.snapshotTitle,
				vvcChangeSetUri: revision["changeSetUrl"],
				componentURI: this.componentURI
			});
			historyHoverPopup.createHoverPopup(div);
		}
		
		return historyHoverPopup;
	},
			
	onThumbnailClick: function(revision){
		var baselineURI = "";
		var ri = this._getFromRevisionsInfoMap(revision);
		if (ri && ri.snapshot) {
			baselineURI = ri.snapshot.resource;
		}
		if (ri && (ri.isModuleBaseline || revision.indexOf("baseline") > 0)) {
			baselineURI = revision;
		}
		if (this.historyViewer && dojo.isFunction(this.historyViewer.updateRevision)){
			var isLast = (ri["http://jazz.net/xmlns/v0.6#successor"] ? false : true);
			this.historyViewer.updateRevision(
				ri.revisionDateLabel,
				revision,
				isLast,
				this.historyViewer.getLatestETag(),
				baselineURI
			);
		}
		
		//changing the URI passes control to the ResourcesPage.showArtifactHistory()
		//if this is a module artifact, we do not want to change the page uri as this will cause the page to 
		// go to the core artifact history instead of keeping the artifact in the module
		if (!this.isModuleArtifact && this.setPageURI) {
			var revision = ri.snapshotRevisionId ? ri.snapshotRevisionId : revision;
			if(ri && (ri.isModuleBaseline || revision.indexOf("baseline") > 0)){
				baselineURI = revision;
				revision = null;
			}
			this.setPageURI(this.resourceURI, revision, baselineURI);
		}
		else{
			this.historyViewer.sidebarArtifactShowRevision(revision, baselineURI);
		}
	},
	
	onThumbnailHover: function(revision){},
	
	destroy: function() {
		this.inherited("destroy", arguments);
	},
 
	getResourceThumbnailsInfo: function(){
		var deferred = new dojo.Deferred();
		var revisionsDeferred = RevisionsCache.getRevisions(this.resourceURI);
		revisionsDeferred.addCallback(this, function(revisionsResponse){
			var deferredArray = [];
			var revisionsParser = dojo.clone(revisionsResponse.doc);
			var descriptionDeferred = com.ibm.rdm.web.model.resources.ResourceManager.getResourceDescribe({
				resourceURI: this.resourceURI,
				revision: this.historyViewer.revision,
				projectSnapshotURI: this.historyViewer.baselineURI,
				ignoreErrors: true
			});
			descriptionDeferred.addCallback(this, function(descriptionResponse){
				this.headResourceDescription = descriptionResponse;
				if (descriptionResponse) {
					deferredArray = this.loadBaselines(descriptionResponse, revisionsParser);
				}
				//artifact is deleted, but we still need to show it in the history viewer
				else {
					var descriptionDeferred2 = com.ibm.rdm.web.model.resources.ResourceManager.getResourceDescribe({
						resourceURI: this.resourceURI,
						revision: this.historyViewer.revision,
						projectSnapshotURI: this.historyViewer.baselineURI
					});
					descriptionDeferred2.addCallback(this, function(descriptionResponse2){
						this.headResourceDescription = descriptionResponse2;
//						deferredArray = this.loadBaselines(descriptionResponse2, revisionsParser);
						revisionsParser.removeRevisionsForDeletedArtifact();
					});
				}
				var dl = new dojo.DeferredList(deferredArray);
				dl.then(function(){
					// callback when we have our project and/or module baselines
					deferred.callback(revisionsParser);
				});
			});
		});
		
		return deferred;
		
		//this._revisionsDeferred.addCallback(this, this.onRevisionsLoaded);
	},
	
	loadBaselines: function(descriptionResponse, revisionsParser){
		var deferredArray = [];
		
		var moduleSnaphotsDeferred = com.ibm.rdm.web.snapshots.ModuleSnapshotServer.getBaselineList(this.owningModule, descriptionResponse.componentURI);
		moduleSnaphotsDeferred.addCallback(this, function(snapshotsResponse){
			revisionsParser.mergeRevisionsWithModuleBaselines(snapshotsResponse.results, this.isModuleArtifact);
			return snapshotsResponse;
		});
		deferredArray.push(moduleSnaphotsDeferred);
		
		return deferredArray;
	}
	
 
 });
 })();
