/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2013, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function() {
dojo.provide("com.ibm.rdm.web.merge.ArtifactMergeWidget");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");

dojo.require("com.ibm.rdm.web.common.ArtifactIcons");
dojo.require("com.ibm.rdm.web.common.DateTimeHelper");
dojo.require("com.ibm.rdm.web.compare.CompareConstants");
dojo.require("com.ibm.rdm.web.compare.compareutils");
dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
dojo.require("com.ibm.rdm.web.merge.AttributeMergeWidget");
dojo.require("com.ibm.rdm.web.merge.ArtifactMergeControlWidget");
dojo.require("com.ibm.rdm.web.compare.ModuleContentCompareWidget");
dojo.require("com.ibm.rdm.web.compare.TabbedArtifactWidget");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.comparevvc.vvccompareutils");
dojo.require("com.ibm.rdm.web.compare.compareLinksUtils");

dojo.requireLocalization("com.ibm.rdm.web.merge", "ArtifactMergeWidget");

var RRMTypes = com.ibm.rdm.web.model.types.RRMTypes,
	ArtifactIcons = com.ibm.rdm.web.common.ArtifactIcons,
	DateTimeHelper = com.ibm.rdm.web.common.DateTimeHelper,
	CompareConstants = com.ibm.rdm.web.compare.CompareConstants,
	CompareUtils = com.ibm.rdm.web.compare.compareutils,
	VVCCompareUtils = com.ibm.rdm.web.comparevvc.vvccompareutils,
	LinksUtils = com.ibm.rdm.web.compare.compareLinksUtils,
	DomUtils = com.ibm.rdm.web.util.domutils;

dojo.declare("com.ibm.rdm.web.merge.ArtifactMergeWidget", [dijit._Widget, dijit._Templated], {
	
	templatePath: dojo.moduleUrl("com.ibm.rdm.web", "merge/templates/ArtifactMergeWidget.html"),
	widgetsInTemplate: true,
	messages: dojo.i18n.getLocalization("com.ibm.rdm.web.merge", "ArtifactMergeWidget"),

	// URIs of the artifacts to be compared.  Including any required ?revision=XXX qualifier.
	_targetURI: null,
	_sourceURI: null,
	// URIs and names of the configurations for the to and from URIs.  Only set for a VVC comparison.
	_targetConfiguration: null,
	_targetConfigurationName: null,
	_sourceConfiguration: null,
	_sourceConfigurationName: null,
	
	// A list of all the changed resources
	_resourceDeltas: null,
	
	// One of the DISPLAY_AS_XXX constants, indicating the type of change being displayed.
	_displayType: null,
	
	// Contains a flag isDeliver or isAccept to indicate the current user operation.
	_operation: null, 
	
	// Attach Points and widgets contained in template
	_headingAreaNode: null,
	_artifactMergeControlWidget: null,
	_attributeMergeWidget: null,
	_moduleMergeWidget: null,
	_nameIdNode: null,
	_loadingIndicator: null,

	// The resources to be compared.
	_targetResource: null,
	_sourceResource: null,
	
	// The types of the resources being compared.  Identical if the type of the resource is the same
	// in both configurations. 
	_targetType: null,
	_sourceType: null,

	// This is either _targetResource or _sourceResource, depending on the type of operation.  Set when the
	// resource(s) are loaded, so that there is a single resource that can be referenced and is guaranteed
	// to be set, reducing the need to constantly look at this._displayType to work out whether _targetResource or
	// sourceResource should be used.
	_primaryResource: null,
	
	// The type of _primaryResource.  Same as one of _targetType or _sourceType, depending on the type of the operation.
	_primaryType: null,
	
	// The Attribute Definitions associated with the _primaryType.
	_primaryAttributeDefinitions: null,
	
	// DOM node representations of the attribute differences.  A hash keyed by attrDefURI.  Each value
	// is an object with fields "label" and "node".
	_attributeDifferences: null,
	_collectionDifferences: null,
	
	// DOM node representations of the tag differences.  It is an object with fields "label", "node",
	// "hasChanged" and "isEmpty".
	_publicTagDifferences: null,
	_privateTagDifferences: null,
	
	// Object containing the Link Type as the string ID, and an array of all the DOM nodes for each link to be shown.
	_linksDifferences: null,
	
	// The Promise relating to a previous call to show() that is still in progress.  Cancelled when a new
	// show request is made to ensure that only the latest show() is active at any time. 
	_inProgressShowPromise: null,
	
	postCreate: function() {
		this.inherited(arguments);
	},
	
	resize: function(/*? integer */width, /*? integer */height) {
		// summary:
		//		Sizes the widget to the given width and height.  Width and height are optional,
		//		allowing this to be used to resize the contained widgets after the content has
		//		been displayed (since the heading area may change in height). 
		// width:
		//		Optional.  If not provided (or zero) the last given width is used.
		// height:
		//		Optional.  If not provided (or zero) the last given height is used.
		
		// for width and height, new new value if provided.  If not provided use previous value.
		width = this._width = width || this._width;
		height = this._height = height || this._height;
		dojo.style(this.domNode, {
			width: width+"px",
			height: height+"px"
		});
		var mainAreaHeight = height - this._artifactMergeControlWidget.domNode.offsetHeight - this._headingAreaNode.offsetHeight;
		this._attributeMergeWidget.resize(width, mainAreaHeight);
		this._moduleMergeWidget.resize(width, mainAreaHeight);
	},
	
	_onMergeControlWidgetResize: function() {
		// summary:
		//		Handler for when the merge control widget resizes
		
		// Resize things contained in this widget to compensate for change in the merge control widget
		this.resize();
	},
	
	show: function(params) {
		// summary:
		//		Loads the given resources and displays them.  Can show an added resource, a
		//		removed resource or a changed resource.
		// displayType:
		//		A CompareConstants.DISPLAY_AS_XXX value, indicating the type of change to display.
		// targetURI:
		//		The URI of the from comparison version (not required for add).
		// targetConfiguration:
		//		The configuration applicable to the from URI (not required for add or non-VVC use).
		// sourceURI:
		//		The URI of the to comparison version (not required for remove).
		// sourceConfiguration:
		//		The configuration applicable to the from URI (not required for remove or non-VVC use).
		this.domNode.style.display = "block";
		this._displayType = params.displayType;
		this._targetURI = params.targetURI;
		this._targetConfiguration = params.targetConfiguration;
		this._targetConfigurationName = params.targetConfigurationName;
		this._sourceURI = params.sourceURI;
		this._sourceConfiguration = params.sourceConfiguration;
		this._sourceConfigurationName = params.sourceConfigurationName;
		this._resourceDeltas = params.resourceDeltas;
		this._operation = params.operation;
		this._mergeStateStore = params.mergeStateStore;
		this._clear();
		if (this._inProgressShowPromise) {
			this._inProgressShowPromise.cancel("SUPERSEDED");
		}		
		this._artifactMergeControlWidget.hide();
		this._showLoading();
		var promise = CompareUtils.getArtifactChanges(params.displayType,
													params.targetURI, params.targetConfiguration,
													params.sourceURI, params.sourceConfiguration,
													params.resourceDeltas)
			.then(dojo.hitch(this, this._parseMergeResources))
			.then(dojo.hitch(this, function() {
				this._artifactMergeControlWidget.show(params.operation.isDeliver, params.resourceDeltas[params.sourceURI], params.mergeStateStore,
						params.sourceConfigurationName, params.targetConfigurationName, this._primaryResource, this._sourceResource, this._targetResource,
						this._sourceConfiguration, this._targetConfiguration,
						this._attributeDifferences, this._linksDifferences);
			}))
			.then(dojo.hitch(this, this._displayHeader))
			.then(dojo.hitch(this, this._showMerge))
			.always(dojo.hitch(this, this._hideLoading))
			.then(dojo.hitch(this, this.resize))
			.always(dojo.hitch(this, this._clearShowPromise));
			
		this._inProgressShowPromise = promise;
		return promise;
	},
	
	hide: function() {
		this.domNode.style.display = "none";
	},
	
	_clear: function() {
		// summary:
		//		Clear the display, ready to display a fresh comparison,
		//		and set internal variables to default values.
		this._nameIdNode.innerHTML = "";
		this._attributeMergeWidget.clear();
		this._moduleMergeWidget.clear();
	},
	
	_parseMergeResources: function(compareResources) {
		// summary:
		//		Loads the compare resources as loaded by CompareUtils.loadResourcesForCompare() into local fields.
		this._targetResource = compareResources.targetResource;
		this._targetType = compareResources.targetType;
		this._sourceResource = compareResources.sourceResource;
		this._sourceType = compareResources.sourceType;
		this._primaryResource = compareResources.primaryResource;
		this._primaryType = compareResources.primaryType;
		this._primaryAttributeDefinitions = compareResources.primaryAttributeDefinitions;
		this._attributeDifferences = compareResources.attributeDifferences;
		this._publicTagDifferences = compareResources.publicTagDifferences;
		this._privateTagDifferences = compareResources.privateTagDifferences;
		this._linksDifferences     = compareResources.linksDifferences;
		this._privateTagDifferences = compareResources.privateTagDifferences;
		this._collectionDifferences = compareResources.collectionDifferences;
		this._teamAreaDifferences = compareResources.teamAreaDifferences;
	},	
	
	_displayHeader: function() {
		// summary:
		//		Populates the widget heading area.
		var componentURI = this._primaryResource.getComponentURI(),
			modifiedOnAttrDefURI = RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.ModifiedOn);
		
		//Show title and Id
		var titleTable = dojo.create("table");
		var titleRow = dojo.create("tr", {}, titleTable);
		var attributeURI = RRMTypes.getWellKnownAttributeURI(componentURI, RRMTypes.WELL_KNOWN_ATTRIBUTE_NAMES.Title);
		var diffInfo = this._attributeDifferences[attributeURI];
		var label = dojo.string.substitute(this.messages.labelDelimiter, [this._primaryResource.getID()]);
		var iconNode = ArtifactIcons.getResourceSmallImage(this._primaryResource);
		dojo.addClass(iconNode, "artifactIcon");
		dojo.create("td", {}, titleRow).appendChild(iconNode);
		dojo.create("td", {}, titleRow).appendChild(document.createTextNode(label));
		if (diffInfo.hasChanged) {
			// Add the attribute merge button menu.
			var attributeMergeButtonMenu = this._artifactMergeControlWidget.getAttributeConflictButtonMenu(attributeURI);
			if (attributeMergeButtonMenu) {
				dojo.create("td", {}, titleRow).appendChild(attributeMergeButtonMenu.domNode);
			}
		}
		// diffInfo.node can be a DOM node or a promise of a DOM node
		dojo.when(diffInfo.node, dojo.hitch(this, function(diffNode){
			dojo.create("td", {}, titleRow).appendChild(diffNode);
			this._nameIdNode.appendChild(titleTable);
		})); 
	},

	_showMerge: function() {
		// summary:
		//		Tell the appropriate contained widget to display the differences.
		
		var resourceFormat = this._primaryResource.getMimeType();
		var projectFormats = RRMTypes.getArtifactFormats(this._primaryResource.getComponentURI());
		
		this._artifactMergeControlWidget.showInfoLine();
		
		if (resourceFormat === projectFormats.Module) {
			// Module
			
			// Hide nodes we do not want displayed, and show the moduleMergeWidget
			if (this._displayType !== CompareConstants.DISPLAY_AS_ADDED && 
				this._displayType !== CompareConstants.DISPLAY_AS_DELETED) {
				// Note that when the display type is DISPLAY_AS_ADDED or DISPLAY_AS_DELETED
				//	only the attributes are shown for modules, and there is no content tab.
				//	We do not create the header bar with the column names, since there are no columns.
				this._artifactMergeControlWidget.hideButtonArea();
				this._artifactMergeControlWidget.hideInfoLine();
			}
			dojo.style(this._attributeMergeWidget.domNode, "display", "none");
			dojo.style(this._moduleMergeWidget.domNode, "display", "block");
			this._moduleMergeWidget.show({
				artifactMergeWidget: this,
				sourceOnLeft: true,
				mergeStateStore: this._mergeStateStore,
				targetResource: this._targetResource,
				targetType: this._targetType,
				targetConfiguration: this._targetConfiguration,
				targetConfigurationName: this._targetConfigurationName,
				sourceResource: this._sourceResource,
				sourceType: this._sourceType,
				sourceConfiguration: this._sourceConfiguration,
				sourceConfigurationName: this._sourceConfigurationName,
				primaryResource: this._primaryResource,
				primaryType: this._primaryType,
				primaryAttributeDefinitions: this._primaryAttributeDefinitions,
				attributeDifferences: this._attributeDifferences,
				publicTagDifferences: this._publicTagDifferences,
				privateTagDifferences: this._privateTagDifferences,
				linksDifferences: this._linksDifferences,
				displayType: this._displayType,
				resourceDeltas: this._resourceDeltas,
				operation: this._operation,
				teamAreaDifferences: this._teamAreaDifferences
			});
		} else {
			// non-Module
			dojo.style(this._moduleMergeWidget.domNode, "display", "none");
			dojo.style(this._attributeMergeWidget.domNode, "display", "block");
			
			var displayType = this._displayType,
				thisDelta = this._resourceDeltas[this._primaryResource.getResourceURI()];
			
			if (VVCCompareUtils.hasOnlyDeltaType(thisDelta, CompareConstants.DELTA_TYPE_LINKS_CHANGE)) {
				displayType = CompareConstants.DISPLAY_AS_LINKS_CHANGED;
			}
			
			LinksUtils.mixinResourceDeltaInfoWithLinksDiff(this._linksDifferences, this._resourceDeltas);
			
			this._attributeMergeWidget.show({
				artifactMergeWidget: this,
				mergeStateStore: this._mergeStateStore,
				targetResource: this._targetResource,
				targetType: this._targetType,
				targetConfiguration: this._targetConfiguration,
				targetConfigurationName: this._targetConfigurationName,
				sourceResource: this._sourceResource,
				sourceType: this._sourceType,
				sourceConfiguration: this._sourceConfiguration,
				sourceConfigurationName: this._sourceConfigurationName,
				primaryResource: this._primaryResource,
				primaryType: this._primaryType,
				primaryAttributeDefinitions: this._primaryAttributeDefinitions,
				attributeDifferences: this._attributeDifferences,
				publicTagDifferences: this._publicTagDifferences,
				privateTagDifferences: this._privateTagDifferences,
				linksDifferences: this._linksDifferences,
				collectionDifferences: this._collectionDifferences,
				displayType: displayType,
				teamAreaDifferences: this._teamAreaDifferences
			});
		}
	},
	
	_clearShowPromise: function() {
		this._inProgressShowPromise = null;
	},
	
	_showLoading: function() {
		dojo.addClass(this._loadingIndicator, "show");
	},
	
	_hideLoading: function() {
		dojo.removeClass(this._loadingIndicator, "show");
	}
	
});
	
})();
