//RunSuiteWizard
/****************************************************************************** 
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2015, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
/*
 * Selection Popup for Script Adapter selection for a Suite
 * Author: vasrivas
 */
dojo.provide("com.ibm.rqm.execution.web.ui.internal.view.suite.RunSuiteWizard");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");
dojo.require("jazz.app.context");
dojo.require("jazz.ui.wizard.WizardStep");

dojo.require("com.ibm.rqm.planning.web.client.internal.Constants");
dojo.require("com.ibm.rqm.execution.web.ui.internal.view.suite.RunSuiteDialog");
dojo.require("com.ibm.rqm.planning.web.ui.internal.view.construction.ExecutionVarsWidget");
dojo.require("com.ibm.rqm.planning.web.ui.internal.view.construction.AttributesCategoriesRequiredFieldsEditor");


(function () {
    var Constants = com.ibm.rqm.planning.web.client.internal.Constants;
    var ExecutionVarsWidget = com.ibm.rqm.planning.web.ui.internal.view.construction.ExecutionVarsWidget;
    var ContentAreas = com.ibm.rqm.planning.web.ui.internal.view.common.InPageContentArea;
    var AttributesCategoriesRequiredFieldsEditor = com.ibm.rqm.planning.web.ui.internal.view.construction.AttributesCategoriesRequiredFieldsEditor;

    var WizardStep_Main = dojo.declare([dijit._Widget, dijit._Templated, jazz.ui.wizard.WizardStep], {
        templateString: '<div></div>',
        stepTitle: Constants.getMessage("labelDefineRecDetails"),
        defaultName: '',
        defaultDesc: '',
        runSuiteWizard: null,

        constructor: function (params) {
            this.params = params;
            this.params.wizardstep = this;
        },

        postCreate: function () {
            this.mainPage = new RunSuiteDialog.RecordDetails(this.params);
            dojo.place(this.mainPage.domNode, this.domNode);
        },

        initialize: function (suite) {
            this.mainPage.initialize(suite);
        },

        canFinish: function () {
            return this.mainPage.stepCanFinish;
        },

        canCancel: function () {
            return this.mainPage.stepCanCancel;
        }
    });

    var WizardStep_CategoriesAttributes = dojo.declare([dijit._Widget, dijit._Templated, jazz.ui.wizard.WizardStep], {
        templateString: '<div></div>',
        stepTitle: Constants.getMessage("labelModifyCategoryAttribute"),
        defaultName: '',
        defaultDesc: '',
        runSuiteWizard: null,
        existingCustomAttributesTSER: null,
        existingCategoriesTSER: null,
        existingCustomAttributesTER: null,
        existingCategoriesTER: null,        
        isRequiredTERs: undefined,
        isRequiredTSERs: undefined,

        constructor: function (params) {
            this.params = params;
        },

        notifyWhenComplete: function () {
            // TODO re-enable when TER editor is ready
            //if ((this.isRequiredTERs !== undefined) && (this.isRequiredTSERs !== undefined)) {
            if (this.isRequiredTSERs !== undefined) {
                this.loadingCompleted();
            }
        },

        postCreate: function () {
            var step = this;
            var tabbedContainer = this.tabbedContainer = new ContentAreas({
                // don't show any scrollable tabs
                tabNoScroll: true
            });
            this.tabbedContainer.initialize({
                tabContainerStyle: "min-height: 100px"
            });
            
            var categoryAttributeEditorForTSER = new AttributesCategoriesRequiredFieldsEditor({
                artifactType: Constants.MODEL_TYPE_SUITE_EXECUTION_RECORD,
                existingCustomAttributes: this.existingCustomAttributesTSER ? this.existingCustomAttributesTSER : null,
                existingCategories: this.existingCategoriesTSER ? this.existingCategoriesTSER : null,
                parent: this,
                onDomNodeInitialized: function (domNode) {
                    step.isRequiredTSERs = domNode && (!categoryAttributeEditorForTSER.validate().status); // TODO maybe this should be part of the requiered editor check
                    if (step.isRequiredTSERs) {
                        step.categoryAttributeEditorForTSER = categoryAttributeEditorForTSER;
                        tabbedContainer.addPage({
                            name: Constants.getMessage("labelSuiteExecutionRecord"),
                            noSelect: true,
                            content: domNode
                        });
                    }
                    step.notifyWhenComplete();
                },
                // If a selection changes on the required category and attribute editor, re-evaluate the valid wizard buttons
                onChange: dojo.hitch(this, step.triggerButtonStateUpdate)
            });

            // TODO the TER editor will require additional server side support before it can be enabled.
//            var categoryAttributeEditorForTER = new AttributesCategoriesRequiredFieldsEditor({
//                artifactType: Constants.MODEL_TYPE_TESTCASE_EXECUTION_RECORD,
//                existingCustomAttributes: this.existingCustomAttributesTER ? this.existingCustomAttributesTER : null,
//                existingCategories: this.existingCategoriesTER ? this.existingCategoriesTER : null,
//                parent: this,
//                onDomNodeInitialized: function (domNode) {
//                    step.isRequiredTERs = domNode && (!categoryAttributeEditorForTER.validate().status);
//                    if (step.isRequiredTERs) {
//                        step.categoryAttributeEditorForTER = categoryAttributeEditorForTER;
//                        tabbedContainer.addPage({
//                            name: Constants.getMessage("labelTestExecutionRecord"),
//                            noSelect: true,
//                            content: domNode
//                        });
//                    }
//                    step.notifyWhenComplete();            
//                },
//                // If a selection changes on the required category and attribute editor, re-evaluate the valid wizard buttons
//                onChange: dojo.hitch(this, step.triggerButtonStateUpdate)
//            });

            dojo.place(this.tabbedContainer.domNode, this.domNode);
        },

        canFinish: function () {
            var isTER_EditorValid = true;
            var isTSER_EditorValid = true;
            if (this.categoryAttributeEditorForTSER) {
                isTSER_EditorValid = this.categoryAttributeEditorForTSER.validate().status;
            }
            if (this.categoryAttributeEditorForTER) {
                isTER_EditorValid = this.categoryAttributeEditorForTER.validate().status;
            }
            return isTER_EditorValid && isTSER_EditorValid;
        }
    });

    var WizardStep_ExecutionVariables = dojo.declare([dijit._Widget, dijit._Templated, jazz.ui.wizard.WizardStep], {
        templateString: '<div></div>',
        stepTitle: Constants.getMessage("labelModExecVars"),
        defaultName: '',
        defaultDesc: '',
        runSuiteWizard: null,
        canFinishOnStep: false,

        constructor: function (params) {
            this.params = params;
        },

        initialize: function (suite) {
            this.variableWidget = new ExecutionVarsWidget({
                artifact: suite,
                artifactType: Constants.TYPE_SUITE,
                inRunDialog: true
            });
            this.variableWidget.initialize();
            dojo.place(this.variableWidget.domNode, this.domNode);
        },

        // Override - This is to handle the event when this step gets the focus.
        focusStep: function () {
        	if (this.variableWidget) {
        		this.variableWidget.refresh();
        	}
    	},
    	
    	// Override - important to return true as Finish button is enabled always for this wizard. 
        canFinish: function() {
        	return true;
        },

    	// Override - important to return true as Cancel button is enabled always for this wizard. 
        canCancel: function() {
        	return true;
        }
    });

    dojo.declare("com.ibm.rqm.execution.web.ui.internal.view.suite.RunSuiteWizard", [dijit._Widget], {
        width: "1000px",
        configContext: null,
        defaultName: '',
        defaultDesc: '',
        allowUseStreamAfterCreate: false,
        showProgressMonitor: false,
        singlePage: true,
        stepMain: null,
        stepCategoriesAttributes: null,
        stepExecutionVariables: null,

        constructor: function (params) {
            this.params = params;
            this.params.runSuiteWizard = this;
        },

        initialize: function (suite) {
            this.stepMain = new WizardStep_Main(this.params);

            var wizardArgs = {
                steps: [this.stepMain],
                width: this.width,
                primaryTitle: Constants.getMessage("labelExeTestSuite"),
                showStepSummary: true,
                processFinish: dojo.hitch(this, "processFinish"),
                customProcessCancel: dojo.hitch(this, "_processWizardCancel"),
                customUpdateControls: this._updateWizardControls
            };
            this.wizard = new jazz.ui.Wizard(wizardArgs);

            // TODO move this out to a subclass of jazz Wizard so others can benefit from the fix.
            // Override original function to patch bug
            this.wizard.contentWidget.removeStep = function (step) {
                if (step._groupNode) {
                    step._groupNode.removeChild(step._stepListDOM);
                } else if (step && step.position > 0) {
                    // The following lines are a work around for foundation WI 200288: 

                    // Fix removing of this step from _steps
                    var index = this._steps.indexOf(step);
                    this._steps.splice(index, 1);

                    // Fix decrementing the step count
                    this._stepList.lastIndex--;

                    // Fix removing the position indicator from the UI
                    this._stepList.wizardSteps.removeChild(this._stepList.wizardSteps.children[index]);

                    // Fix removing the step from the DOM
                    this.wizardPanelArea.removeChild(step._stepDOM);
                }
            };
            dojo.addClass(this.wizard.contentWidget.wizardFinishButton, "primary-button");
            dojo.addClass(this.wizard.contentWidget.wizardCancelButton, "non-primary-button");
            dojo.addClass(this.wizard.contentWidget.wizardNextButton, "non-primary-button");
            dojo.addClass(this.wizard.contentWidget.wizardPreviousButton, "non-primary-button");

            this.wizard.viewBox.setIcon(this.dialogIconPath);
            this.wizard.viewBox.setClosable(true);

            this.domNode = this.wizard.domNode;

            this.stepMain.initialize(suite);
        },

        // Will create and add the category attribute wizard page only when there are required fields which are not set
        // Otherwise no page will be added or if there is an existing page it will be removed
        enableStepCategoriesAttributesWhenRequiredFields: function (suiteExecutionRecord) {
            if (this.stepCategoriesAttributes) {
                if (this.curentSuiteRecord !== suiteExecutionRecord) {
                    this.wizard.removeStep(this.stepCategoriesAttributes);
                    this.stepCategoriesAttributes = null;
                } else {
                    return;
                }
            }

            var customAttributesTSER = null;
            var categoriesTSER = null;
            if (suiteExecutionRecord) {
                customAttributesTSER = suiteExecutionRecord.customAttributes ? suiteExecutionRecord.customAttributes : null;
                categoriesTSER = suiteExecutionRecord.categories ? suiteExecutionRecord.categories : null;
            }
            this.curentSuiteRecord = suiteExecutionRecord;
            var stepCategoriesAttributes = new WizardStep_CategoriesAttributes({
                existingCustomAttributesTSER: customAttributesTSER,
                existingCategoriesTSER: categoriesTSER,
                // TODO fetch unique lists of attributes and categories for all TER's that belong to the TSER
                existingCustomAttributesTER: null,
                existingCategoriesTER: null,                
                runSuiteWizard: this,
                loadingCompleted: dojo.hitch(this, function () {
                    if (stepCategoriesAttributes.isRequiredTERs || stepCategoriesAttributes.isRequiredTSERs) {
                        this.stepCategoriesAttributes = stepCategoriesAttributes;
                        this.wizard.addStep(this.stepCategoriesAttributes);
                        this.wizard.contentWidget.updateNavigationControls();
                    }
                })
            });
        },

        createExecutionVariablesStep: function (suite) {
            this.stepExecutionVariables = new WizardStep_ExecutionVariables();
            this.wizard.addStep(this.stepExecutionVariables);
            this.stepExecutionVariables.initialize(suite);
            this.wizard.contentWidget.updateNavigationControls();
        },

        removeExecutionVariablesStep: function (suite) {
            this.wizard.removeStep(this.stepExecutionVariables);
            this.stepExecutionVariables = null;
            this.wizard.contentWidget.updateNavigationControls();
        },
        
         // This function used to get the content of variableWidget
         getExecutionVariablesMap: function () {
        	if (this.stepExecutionVariables && this.stepExecutionVariables.variableWidget) {
        		return this.stepExecutionVariables.variableWidget.getContent();
        	}
            return null;
        },  

        startup: function () {
            this.wizard.startup();
        },
        
		setCSHelpId: function(helpId){	
            var csHelpAttachPoint = this.wizard.viewBox._headerSecondary;
	        if (csHelpAttachPoint) {
	            dojox.xml.parser.removeChildren(csHelpAttachPoint);
	            if (helpId) {
	                csHelpAttachPoint.style.display = "inline";
	                var args = {
	                    'contextID': helpId
	                };
					if(this.cshelp)
						this.cshelp.destroy();
	                this.cshelp = new com.ibm.asq.common.web.ui.internal.util.CSHelp(args);
	                csHelpAttachPoint.appendChild(this.cshelp.domNode);
	            } else {
	                csHelpAttachPoint.style.display = "none";
	            }
	        }
	    },        

        _processWizardCancel: function () {
            this.wizard.close();
        },

        // Enforce all wizard pages to be valid before we can click finish
        _updateWizardControls: function () {
            var isAnyStepInvalid = dojo.some(this._steps, function (step) {
                return !step.canFinish();
            });
            this.wizardFinishButton.disabled = isAnyStepInvalid;
        },

        processFinish: function () {
            this.stepMain.mainPage._onRunClick();
        },

        uninitialize: function () {
            this.wizard.close();
        }
    });
})();